"""
Created on 7 Aug 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://stackoverflow.com/questions/403421/how-do-i-sort-a-list-of-objects-based-on-an-attribute-of-the-objects

example document:
{"endpoint": "test1", "client": "MyOrg", "period": "2023-08-22", "queries": 99, "invocations": 99, "characters": 495}
"""

import re

from abc import ABC
from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class ClientTrafficLocus(ABC, JSONable):
    """
    classdocs
    """

    @classmethod
    def is_valid_period(cls, period):
        try:
            match = re.match(r'^2\d{3}(-[01]\d(-[0123]\d)?)?$', period)
            return match is not None

        except TypeError:
            return False


    @classmethod
    def now(cls):
        return str(LocalizedDatetime.now().utc().datetime.date())


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, endpoint, client, period):
        """
        Constructor
        """
        self.__endpoint = endpoint                              # string or None
        self.__client = client                                  # string (or array of) email or int OrgID
        self.__period = period                                  # string (part-)date, e.g. 2023, 2023-08, 2023-08-31


    def __eq__(self, other):
        try:
            return self.endpoint == other.endpoint and self.client == other.client \
                and self.period == other.period

        except (TypeError, AttributeError):
            return False


    def __lt__(self, other):
        if self.endpoint.lower() < other.endpoint.lower():
            return True

        if self.endpoint.lower() > other.endpoint.lower():
            return False

        if self.client.lower() < other.client.lower():
            return True

        if self.client.lower() > other.client.lower():
            return False

        if self.period < other.period:
            return True

        return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def endpoint(self):
        return self.__endpoint


    @property
    def client(self):
        return self.__client


    @property
    def period(self):
        return self.__period


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ClientTrafficLocus:{endpoint:%s, client:%s, period:%s}" % \
            (self.endpoint, self.client, self.period)


# --------------------------------------------------------------------------------------------------------------------

class ClientTrafficReport(ClientTrafficLocus):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def client_aggregations(cls, aggregation_period, reports):
        aggregations = {}

        for report in reports:
            key = '+'.join((report.endpoint, report.client))

            if key in aggregations:
                aggregations[key] += report
            else:
                aggregations[key] = cls(report.endpoint, report.client, aggregation_period,
                                        report.queries, report.invocations, report.characters)

        return tuple(aggregations.values())


    @classmethod
    def organisation_totals(cls, org_label, reports):
        totals = {}

        for report in reports:
            key = '+'.join((report.endpoint, report.period))

            if key in totals:
                totals[key] += report
            else:
                totals[key] = cls(report.endpoint, org_label, report.period,
                                  report.queries, report.invocations, report.characters)

        return tuple(totals.values())


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        endpoint = jdict.get('endpoint')
        client = jdict.get('client')
        period = jdict.get('period')

        queries = jdict.get('queries')
        invocations = jdict.get('invocations')
        characters = jdict.get('characters')

        return cls(endpoint, client, period, queries, invocations, characters)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, endpoint, client, period, queries, invocations, characters):
        """
        Constructor
        """
        super().__init__(endpoint, client, period)

        self.__queries = int(queries)                           # int
        self.__invocations = int(invocations)                   # int
        self.__characters = int(characters)                     # int


    def __add__(self, other):
        return ClientTrafficReport(self.endpoint, self.client, self.period,
                                   self.__queries + other.queries,
                                   self.__invocations + other.invocations,
                                   self.__characters + other.characters)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['endpoint'] = self.endpoint
        jdict['client'] = self.client
        jdict['period'] = self.period

        jdict['queries'] = self.queries
        jdict['invocations'] = self.invocations
        jdict['characters'] = self.characters

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def queries(self):
        return self.__queries


    @property
    def invocations(self):
        return self.__invocations


    @property
    def characters(self):
        return self.__characters


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        class_name = self.__class__.__name__

        return class_name + ":{endpoint:%s, client:%s, period:%s, queries:%s, invocations:%s, characters:%s}" % \
            (self.endpoint, self.client, self.period, self.queries, self.invocations, self.characters)



"""
Created on 22 Dec 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

an ML model group configuration for gases inference

example JSON:
{"uds-path": "pipes/lambda-gas-model.uds", "model-interface": "vE", "model-map": "oM.2"}
"""

from scs_core.gas.afe_calib import AFECalib

from scs_core.model.catalogue.model_compendium_group import ModelCompendiumGroup
from scs_core.model.gas.gas_baseline import GasBaseline
from scs_core.model.gas.gas_inference_client import GasInferenceClient
from scs_core.model.gas.vcal_baseline import VCalBaseline
from scs_core.model.model_conf import ModelConf

from scs_core.sync.schedule import ScheduleItem


# --------------------------------------------------------------------------------------------------------------------

class GasModelConf(ModelConf):
    """
    classdocs
    """

    __FILENAME = "gas_model_conf.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    __INTERFACES = ('s1', 'vB', 'vB2', 'vE')

    @classmethod
    def interfaces(cls):
        return cls.__INTERFACES


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, uds_path, model_interface, model_map):
        """
        Constructor
        """
        super().__init__(uds_path, model_interface, model_map)


    # ----------------------------------------------------------------------------------------------------------------

    def client(self, host, socket, schedule_item: ScheduleItem) -> GasInferenceClient:
        # s1 (obsolete)...
        if self.model_interface == 's1':
            from scs_core.model.gas.s1.s1_gas_inference_client import S1GasInferenceClient

            afe_calib = AFECalib.load(host)

            return S1GasInferenceClient.construct(socket, self.abs_uds_path(host), schedule_item, afe_calib)

        # vB (document v1.0 and v2.0)...
        if self.model_interface == 'vB' or self.model_interface == 'vB2':
            from scs_core.model.gas.vB.vb_gas_inference_client import VBGasInferenceClient

            return VBGasInferenceClient.construct(socket, self.abs_uds_path(host), schedule_item)

        # vE (document v2.0)...
        if self.model_interface == 'vE':
            from scs_core.model.gas.vE.ve_gas_inference_client import VEGasInferenceClient

            vcal_baseline = VCalBaseline.load(host, skeleton=True)
            gas_baseline = GasBaseline.load(host, skeleton=True)
            model_compendium_group = ModelCompendiumGroup.retrieve(self.model_map.gas_compendium_group)

            return VEGasInferenceClient.construct(socket, self.abs_uds_path(host), schedule_item, vcal_baseline,
                                                  gas_baseline, model_compendium_group)

        raise ValueError(self.model_interface)



"""
Created on 28 Oct 2020

@author: Jade Page (jade.page@southcoastscience.com)
"""

import time

from collections import OrderedDict
from multiprocessing import Manager

from scs_core.email.email_queue import EmailQueue

from scs_core.sync.interval_timer import IntervalTimer
from scs_core.sync.synchronised_process import SynchronisedProcess

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class EmailQueueManager(SynchronisedProcess):

    __WAIT_PERIOD = 10  # seconds
    __MAX_NUM_RETRIES = 5


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, email_client):
        """
        Constructor
        """
        self.__logger = Logging.getLogger()
        self.__logging_specification = Logging.specification()

        manager = Manager()

        SynchronisedProcess.__init__(self, value=manager.list())

        self.__email_client = email_client


    # ----------------------------------------------------------------------------------------------------------------

    def start(self):
        self.__email_client.open_server()
        super().start()


    def stop(self):
        time.sleep(self.__WAIT_PERIOD * 2)
        super().stop()


    def run(self):
        Logging.replicate(self.__logging_specification)

        try:
            timer = IntervalTimer(10)

            while timer.true():
                k, v = None, None
                if self._value is not None:
                    with self._lock:
                        queue = EmailQueue.construct_from_jdict(OrderedDict(self._value))
                        if queue is not None:
                            k, v = queue.pop_next()
                    if k is not None and v is not None:
                        if self.__email_client.send_mime_email(v, k):
                            self.set_queue(queue)

                time.sleep(self.__WAIT_PERIOD)

        except (ConnectionError, KeyboardInterrupt, SystemExit):
            pass


    def set_queue(self, queue):
        with self._lock:
            queue.as_list(self._value)

        return True


    def get_queue(self):
        with self._lock:
            queue = EmailQueue.construct_from_jdict(OrderedDict(self._value))

        return queue


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "EmailQueueManager:{queue:%s}" % self.get_queue()



"""
Created on 19 Apr 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

separated from client to remove dependency on requests package
"""

from abc import ABC, abstractmethod
from http import HTTPStatus

from scs_core.data.json import JSONable, JSONify


# --------------------------------------------------------------------------------------------------------------------

class APIResponse(ABC, JSONable):
    """
    classdocs
    """

    __CORS_HEADERS = {                                      # Cross-Origin Resource Sharing
        'Access-Control-Allow-Origin': '*',
        'Access-Control-Allow-Credentials': True
    }

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def construct_from_jdict(cls, jdict):
        pass


    # ----------------------------------------------------------------------------------------------------------------
    # server...

    def as_http(self, status=HTTPStatus.OK, cors=False):
        jdict = {
            'statusCode': int(status),
            'body': JSONify.dumps(self)
        }

        if cors:
            jdict['headers'] = self.__CORS_HEADERS

        return jdict


    # ----------------------------------------------------------------------------------------------------------------
    # client...

    def start(self):
        return None


    @abstractmethod
    def next_params(self, params):
        pass


    @property
    @abstractmethod
    def next_url(self):
        pass



"""
Created on 25 Jul 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://howchoo.com/g/ywi5m2vkodk/working-with-datetime-objects-and-timezones-in-python

example JSON:
{"sample-period": {"type": "diurnal", "start": "23:00:00", "end": "08:00:00", "timezone": "Europe/London"},
"aggregation-period": {"type": "recurring", "interval": 5, "units": "M", "timezone": "Europe/London"},
"minimums": {"CO": 300, "CO2": 420, "H2S": 5, "NO": 10, "NO2": 10, "Ox": 50, "SO2": 5, "VOC": 250}}
"""

import os

from collections import OrderedDict

from scs_core.data.json import MultiPersistentJSONable
from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.diurnal_period import DiurnalPeriod
from scs_core.data.recurring_period import RecurringPeriod
from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class BaselineConf(MultiPersistentJSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    __SUPPORTED_GASES = ('CO', 'CO2', 'H2S', 'NO', 'NO2', 'Ox', 'SO2', 'VOC', 'VOCe')

    @classmethod
    def supported_gases(cls):
        return cls.__SUPPORTED_GASES


    # ----------------------------------------------------------------------------------------------------------------

    __FILENAME = "baseline_conf.json"

    @classmethod
    def persistence_location(cls, name):
        filename = cls.__FILENAME if name is None else '_'.join((name, cls.__FILENAME))

        return os.path.join(cls.conf_dir(), 'baseline_conf'), filename


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, name=None, skeleton=False):
        if not jdict:
            return cls(None, None, None, {}) if skeleton else None

        sample_period = DiurnalPeriod.construct_from_jdict(jdict.get('sample-period'))
        aggregation_period = RecurringPeriod.construct_from_jdict(jdict.get('aggregation-period'))
        minimums = jdict.get('minimums')

        return cls(name, sample_period, aggregation_period, minimums)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, name, sample_period, aggregation_period, minimums):
        """
        Constructor
        """
        super().__init__(name)

        self.__sample_period = sample_period                    # DiurnalPeriod
        self.__aggregation_period = aggregation_period          # RecurringPeriod
        self.__minimums = minimums                              # dict of string: int


    # ----------------------------------------------------------------------------------------------------------------

    def duplicate(self, new_name):
        sample_period = self.sample_period
        aggregation_period = self.aggregation_period
        minimums = self.minimums

        return BaselineConf(new_name, sample_period, aggregation_period, minimums)


    # ----------------------------------------------------------------------------------------------------------------

    def start_datetime(self, origin: LocalizedDatetime):
        return self.sample_period.start_datetime(origin)


    def end_datetime(self, origin: LocalizedDatetime):
        return self.sample_period.end_datetime(origin)


    def expected_data_points(self, start, end):
        points_per_hour = 60 / self.aggregation_period.interval
        period = end - start

        return int(points_per_hour * period.hours)


    def checkpoint(self):
        return self.aggregation_period.checkpoint()


    # ----------------------------------------------------------------------------------------------------------------

    def set_minimum(self, gas, value):
        self.__minimums[gas] = value


    def remove_minimum(self, gas):
        try:
            del self.__minimums[gas]
        except KeyError:
            pass


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def sample_period(self):
        return self.__sample_period


    @property
    def aggregation_period(self):
        return self.__aggregation_period

    @property
    def timezone(self):
        return None if self.sample_period is None else self.sample_period.timezone


    @property
    def gases(self):
        return set(sorted(self.__minimums.keys()))


    def minimum(self, gas):
        return self.__minimums[gas]


    @property
    def minimums(self):
        return OrderedDict(sorted(self.__minimums.items()))


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['sample-period'] = self.sample_period
        jdict['aggregation-period'] = self.aggregation_period
        jdict['minimums'] = self.minimums

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "BaselineConf:{name:%s, sample_period:%s, aggregation_period:%s, minimums:%s}" %  \
               (self.name, self.sample_period, self.aggregation_period, Str.collection(self.minimums))



"""
Created on 17 Jan 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import json

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint
from scs_core.aws.security.cognito_user import CognitoUserIdentity

from scs_core.aws.security.organisation import Organisation, OrganisationPathRoot, OrganisationUser, \
    OrganisationUserPath, OrganisationDevice

from scs_core.data.json import JSONify


# --------------------------------------------------------------------------------------------------------------------

class DevEndpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('OrgDevAPI/DeviceOrganisationManager',
                   'https://6fs21ckyci.execute-api.us-west-2.amazonaws.com/default/DeviceOrganisationManager')


# --------------------------------------------------------------------------------------------------------------------

class ExecEndpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('OrgExecAPI/OrganisationExecutive',
                   'https://19tm2j6odj.execute-api.us-west-2.amazonaws.com/default/OrganisationExecutive')


# --------------------------------------------------------------------------------------------------------------------

class MgrEndpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('OrgMgrAPI/OrganisationManager',
                   'https://04h65m94o8.execute-api.us-west-2.amazonaws.com/default/OrganisationManager')


# --------------------------------------------------------------------------------------------------------------------

class OrganisationManager(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------
    # Organisation...

    def find_organisations(self, token):
        url = MgrEndpoint.url('organisation')

        response = requests.get(url, headers=self._token_headers(token))
        self._check_response(response)

        return tuple(Organisation.construct_from_jdict(jdict) for jdict in response.json())


    def find_child_organisations(self, token, label):
        url = MgrEndpoint.url('organisation')
        payload = JSONify.dumps({"ParentLabel": label})

        response = requests.get(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)

        return tuple(Organisation.construct_from_jdict(jdict) for jdict in response.json())


    def get_organisation(self, token, id):
        url = MgrEndpoint.url('organisation')
        payload = JSONify.dumps({"ID": id})

        response = requests.get(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)

        return Organisation.construct_from_jdict(response.json())


    def get_organisation_by_label(self, token, label):
        url = MgrEndpoint.url('organisation')
        payload = JSONify.dumps({"Label": label})

        response = requests.get(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)

        return Organisation.construct_from_jdict(response.json())


    def insert_organisation(self, token, organisation):
        url = ExecEndpoint.url('organisation')
        payload = JSONify.dumps(organisation)

        response = requests.post(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)

        return Organisation.construct_from_jdict(response.json())


    def update_organisation(self, token, organisation):
        url = MgrEndpoint.url('organisation')
        payload = JSONify.dumps(organisation)

        response = requests.patch(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)


    def delete_organisation(self, token, org_id):
        url = ExecEndpoint.url('organisation')
        payload = JSONify.dumps({"OrgID": org_id})

        response = requests.delete(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)


    # ----------------------------------------------------------------------------------------------------------------
    # OrganisationPathRoot...

    def find_oprs(self, token, org_id=None):
        url = MgrEndpoint.url('opr')
        payload = {}

        if org_id:
            payload['OrgID'] = org_id

        response = requests.get(url, headers=self._token_headers(token), data=JSONify.dumps(payload))
        self._check_response(response)

        return tuple(OrganisationPathRoot.construct_from_jdict(jdict) for jdict in response.json())


    def get_opr_by_path_root(self, token, path_root):
        url = MgrEndpoint.url('opr')
        payload = JSONify.dumps({"PathRoot": path_root})

        response = requests.get(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)

        return OrganisationPathRoot.construct_from_jdict(response.json())


    def insert_opr(self, token, opr):
        url = ExecEndpoint.url('opr')
        payload = JSONify.dumps(opr)

        response = requests.post(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)

        return OrganisationPathRoot.construct_from_jdict(response.json())


    def delete_opr(self, token, opr_id):
        url = ExecEndpoint.url('opr')
        payload = JSONify.dumps({"OPRID": opr_id})

        response = requests.delete(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)


    # ----------------------------------------------------------------------------------------------------------------
    # OrganisationUser...

    def find_users(self, token):
        url = MgrEndpoint.url('user')

        response = requests.get(url, headers=self._token_headers(token))
        self._check_response(response)

        return tuple(OrganisationUser.construct_from_jdict(jdict) for jdict in response.json())


    def find_users_by_organisation(self, token, org_id):
        url = MgrEndpoint.url('user')
        payload = JSONify.dumps({"OrgID": org_id})

        response = requests.get(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)

        return tuple(OrganisationUser.construct_from_jdict(jdict) for jdict in response.json())


    def find_cognito_users_by_organisation(self, token, org_id):
        url = MgrEndpoint.url('cognito-user')
        payload = JSONify.dumps({"OrgID": org_id})

        response = requests.get(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)

        return tuple(CognitoUserIdentity.construct_from_jdict(jdict) for jdict in response.json())


    def find_users_by_username(self, token, username):
        url = MgrEndpoint.url('user')
        payload = JSONify.dumps({"Username": username})

        response = requests.get(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)

        return tuple(OrganisationUser.construct_from_jdict(jdict) for jdict in response.json())


    def get_user(self, token, username, org_id):
        url = MgrEndpoint.url('user')
        payload = JSONify.dumps({"Username": username, "OrgID": org_id})

        response = requests.get(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)

        return OrganisationUser.construct_from_jdict(response.json())


    def assert_user(self, token, user):
        url = MgrEndpoint.url('user')
        payload = JSONify.dumps(user)

        response = requests.post(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)


    def delete_user(self, token, username, org_id):
        url = ExecEndpoint.url('user')
        payload = JSONify.dumps({"Username": username, "OrgID": org_id})

        response = requests.delete(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)


    # ----------------------------------------------------------------------------------------------------------------
    # OrganisationUserPath...

    def find_oups(self, token, username=None, opr_id=None):
        url = MgrEndpoint.url('oup')
        payload = {}

        if username:
            payload['Username'] = username

        if opr_id:
            payload['OPRID'] = opr_id

        response = requests.get(url, headers=self._token_headers(token), data=JSONify.dumps(payload))
        self._check_response(response)

        return tuple(OrganisationUserPath.construct_from_jdict(jdict) for jdict in response.json())


    def assert_oup(self, token, oup):
        url = MgrEndpoint.url('oup')
        payload = JSONify.dumps(oup)

        response = requests.post(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)


    def delete_oup(self, token, oup):
        url = MgrEndpoint.url('oup')
        payload = JSONify.dumps(oup)

        response = requests.delete(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)


    # ----------------------------------------------------------------------------------------------------------------
    # OrganisationDevice...

    def find_devices(self, token):
        url = MgrEndpoint.url('device')

        response = requests.get(url, headers=self._token_headers(token))
        self._check_response(response)

        return tuple(OrganisationDevice.construct_from_jdict(jdict) for jdict in response.json())


    def find_devices_by_tag(self, token, device_tag):
        url = MgrEndpoint.url('device')
        payload = JSONify.dumps({"DeviceTag": device_tag})

        response = requests.get(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)

        return tuple(OrganisationDevice.construct_from_jdict(jdict) for jdict in response.json())


    def find_devices_by_organisation(self, token, org_id):
        url = MgrEndpoint.url('device')
        payload = JSONify.dumps({"OrgID": org_id})

        response = requests.get(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)

        return tuple(OrganisationDevice.construct_from_jdict(jdict) for jdict in response.json())


    def assert_device(self, token, device):
        url = ExecEndpoint.url('device')
        payload = JSONify.dumps(device)

        response = requests.post(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)


    def delete_device(self, token, device_tag):
        url = ExecEndpoint.url('device')
        payload = JSONify.dumps({"DeviceTag": device_tag})

        response = requests.delete(url, headers=self._token_headers(token), data=payload)
        self._check_response(response)


# --------------------------------------------------------------------------------------------------------------------

class DeviceOrganisationManager(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------
    # OrganisationDevice...

    def location_path_in_use(self, token, location_path):
        payload = JSONify.dumps({"LocationPath": location_path})

        response = requests.get(DevEndpoint.url(), headers=self._token_headers(token), data=payload)
        self._check_response(response)

        return json.loads(response.json())



"""
Created on 11 Jul 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

settings for OPCMonitor

example JSON:
{"model": "N3", "sample-period": 10, "restart-on-zeroes": true, "power-saving": false,
"custom-dev-path": "/dev/spi/by-connector/H3"}
"""

from collections import OrderedDict

from scs_core.data.json import MultiPersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class OPCConf(MultiPersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "opc_conf.json"

    @classmethod
    def persistence_location(cls, name):
        filename = cls.__FILENAME if name is None else '_'.join((name, cls.__FILENAME))

        return cls.conf_dir(), filename


    @classmethod
    def load(cls, manager, name=None, encryption_key=None, skeleton=False):
        conf = super().load(manager, name=name, encryption_key=encryption_key, skeleton=skeleton)

        if conf:
            conf.__default_dev_path = manager.opc_spi_dev_path()

        return conf


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, name=None, skeleton=False):
        if not jdict:
            return None

        model = jdict.get('model')
        sample_period = jdict.get('sample-period')
        restart_on_zeroes = jdict.get('restart-on-zeroes', True)
        power_saving = jdict.get('power-saving')

        custom_dev_path = jdict.get('custom-dev-path')

        return cls(model, sample_period, restart_on_zeroes, power_saving, custom_dev_path, name=name)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, model, sample_period, restart_on_zeroes, power_saving, custom_dev_path, name=None):
        """
        Constructor
        """
        super().__init__(name)

        self.__model = model                                        # string
        self.__sample_period = int(sample_period)                   # int
        self.__restart_on_zeroes = bool(restart_on_zeroes)          # bool
        self.__power_saving = bool(power_saving)                    # bool

        self.__default_dev_path = None                              # string
        self.__custom_dev_path = custom_dev_path                    # string


    def __eq__(self, other):                            # ignore name
        try:
            return self.model == other.model and self.sample_period == other.sample_period and \
                   self.restart_on_zeroes == other.restart_on_zeroes and self.power_saving == other.power_saving and \
                   self.custom_dev_path == other.custom_dev_path

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def opc_monitor(self, manager, interface):
        return None


    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def opc(self, interface):
        return None


    # noinspection PyMethodMayBeStatic
    def uses_spi(self):
        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def model(self):
        return self.__model


    @property
    def sample_period(self):
        return self.__sample_period


    @property
    def power_saving(self):
        return self.__power_saving


    @property
    def restart_on_zeroes(self):
        return self.__restart_on_zeroes


    @property
    def dev_path(self):
        return self.default_dev_path if self.custom_dev_path is None else self.custom_dev_path


    @property
    def default_dev_path(self):
        return self.__default_dev_path


    @property
    def custom_dev_path(self):
        return self.__custom_dev_path


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['model'] = self.model
        jdict['sample-period'] = self.sample_period
        jdict['restart-on-zeroes'] = self.restart_on_zeroes
        jdict['power-saving'] = self.power_saving

        if self.custom_dev_path is not None:
            jdict['custom-dev-path'] = self.custom_dev_path

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "OPCConf(core):{name:%s, model:%s, sample_period:%s, restart_on_zeroes:%s, power_saving:%s, " \
               "default_dev_path:%s, custom_dev_path:%s}" %  \
               (self.name, self.model, self.sample_period, self.restart_on_zeroes, self.power_saving,
                self.default_dev_path, self.custom_dev_path)



"""
Created on 2 Nov 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

one cell of a statistical model, with one independent variable and one or more dependent variables

https://en.wikipedia.org/wiki/Dependent_and_independent_variables

example:
{"domain": "70.0 - 75.0", "praxis": {"climate": {"val": {"hmd": {"min": 70.0, "avg": 72.6, "max": 74.9}}}},
"error": {"pm1": {"avg": 2.648, "stdev": 2.458}, "pm2p5": {"avg": 2.992, "stdev": 2.652},
"pm10": {"avg": 2.77, "stdev": 2.456}}, "samples": 2307}
"""

from collections import OrderedDict
from statistics import mean, stdev

from scs_core.data.json import JSONable
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

class ModelDelta(JSONable):
    """
    classdocs
    """

    DEFAULT_INDEPENDENT_PREC =      1
    DEFAULT_DEPENDENT_PREC =        3

    __STDEV_PRECISION =             3

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, domain_min, domain_max, ind_name, ind_prec, dep_names, dep_prec):
        dependents = OrderedDict()
        for dep_name in dep_names:
            dependents[dep_name] = []

        return ModelDelta(domain_min, domain_max, ind_name, [], ind_prec, dependents, dep_prec)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, domain_min, domain_max, ind_name, ind_values, ind_prec, dependents, dep_prec):
        """
        Constructor
        """
        self.__domain_min = domain_min                          # float
        self.__domain_max = domain_max                          # float

        self.__ind_name = ind_name                              # string
        self.__ind_values = ind_values                          # array of float
        self.__ind_prec = ind_prec                              # int precision

        self.__dependents = dependents                          # dict of string: array of float
        self.__dep_prec = dep_prec                              # int precision


    def __len__(self):
        return len(self.__ind_values)


    # ----------------------------------------------------------------------------------------------------------------

    def append(self, ind_value, dependents):
        if len(dependents) != len(self.__dependents):
            raise IndexError("number of dependents given:%d required:%d" % (len(dependents), len(self.__dependents)))

        self.__ind_values.append(ind_value)

        for key in dependents.keys():
            self.__dependents[key].append(dependents[key])


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = PathDict()

        jdict.append('domain', self.domain())
        jdict.append('domain-midpoint', self.domain_midpoint())

        jdict.append(self.ind_name + '.min', self.ind_min())
        jdict.append(self.ind_name + '.avg', self.ind_avg())
        jdict.append(self.ind_name + '.max', self.ind_max())

        for name in self.dep_names:
            jdict.append(name + '.avg', self.dep_avg(name))
            jdict.append(name + '.stdev', self.dep_stdev(name))

        jdict.append('samples', len(self))

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def ind_min(self):
        if not self.__ind_values:
            return None

        value = min(self.__ind_values)
        return round(value, self.__ind_prec)


    def ind_avg(self):
        if not self.__ind_values:
            return None

        value = sum(self.__ind_values) / len(self)
        return round(value, self.__ind_prec)


    def ind_max(self):
        if not self.__ind_values:
            return None

        value = max(self.__ind_values)
        return round(value, self.__ind_prec)


    def dep_avg(self, name):
        if not self.__dependents[name]:
            return None

        value = sum(self.__dependents[name]) / len(self)
        return round(value, self.__dep_prec)


    def dep_stdev(self, name):
        if not self.__dependents[name] or len(self.__dependents[name]) == 1:
            return None

        value = stdev(self.__dependents[name])
        return round(value, self.__STDEV_PRECISION)


    # ----------------------------------------------------------------------------------------------------------------

    def domain(self):
        return ' - '.join((str(round(self.domain_min, 1)), str(round(self.domain_max, 1))))


    def domain_midpoint(self):
        return round(mean((self.domain_min, self.domain_max)), 1)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def domain_min(self):
        return self.__domain_min


    @property
    def domain_max(self):
        return self.__domain_max


    @property
    def ind_name(self):
        return self.__ind_name


    @property
    def dep_names(self):
        return tuple(self.__dependents.keys())


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ModelDelta:{domain:%s, ind_name:%s, ind_prec:%s, dependents:%s, dep_prec:%s, length:%d}" % \
               (self.domain(), self.ind_name, self.__ind_prec, self.dep_names, self.__dep_prec, len(self))



"""
Created on 27 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import json
import re

from collections import OrderedDict
from copy import deepcopy

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class PathDict(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def union(*pairs):
        union = PathDict()

        for name, value in pairs:
            if value is not None:
                union.append(name, value)

        return union


    @staticmethod
    def sub_path_includes_path(sub_path, path):
        sub_path_nodes = [node[0] for node in re.findall(r'([^.:]+)([.:]*)', sub_path)]
        path_nodes = [node[0] for node in re.findall(r'([^.:]+)([.:]*)', path)]

        for i in range(len(sub_path_nodes)):
            try:
                if sub_path_nodes[i] != path_nodes[i]:
                    return False
            except IndexError:
                return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jstr(cls, jstr, sort_paths=()):
        try:
            jdict = json.loads(jstr)
        except ValueError:
            return None

        return cls(dictionary=jdict, sort_paths=sort_paths)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, dictionary=None, sort_paths=()):
        """
        Constructor
        """
        self.__dictionary = OrderedDict() if dictionary is None else dictionary
        self.__sort_paths = sort_paths


    def __len__(self):
        return len(self.__dictionary)


    def __lt__(self, other):
        for sort_path in self.__sort_paths:
            value = self.node(sub_path=sort_path)
            other_value = other.node(sub_path=sort_path)

            if value < other_value:
                return True

            if value > other_value:
                return False

        return False


    # -----------------------------------------------------------------------------------------------------------------
    # source...

    # Tests whether a leaf node path is present...

    def has_path(self, path):
        return path in self.paths()


    # Tests whether a leaf node or internal path is present...

    def has_sub_path(self, sub_path=None):
        if sub_path is None:
            return True

        try:
            self.node(sub_path)
            return True

        except (KeyError, TypeError):
            return False


    # Returns all the leaf node paths...

    def paths(self, sub_path=None):
        node = self if sub_path is None else PathDict(self.node(sub_path))

        return node.__paths(node.__dictionary, sub_path)


    # Returns a leaf node or internal node...

    def node(self, sub_path=None):
        if sub_path is None:
            return self.__dictionary

        try:
            return self.__node(self.__dictionary, re.split(r"[.:]", sub_path))

        except (KeyError, TypeError):
            raise KeyError(sub_path)


    # Appends or replaces value at sub_path...

    def append(self, sub_path, value):
        nodes = re.findall('([^.:]+)([.:]*)', sub_path)

        try:
            self.__append(self.__dictionary, nodes, value)

        except (KeyError, TypeError):
            raise KeyError(sub_path)


    # Sets an internal node to the given value...

    def set_node(self, sub_path, value):
        try:
            return self.__set_node(self.__dictionary, re.split(r"[.:]", sub_path), value)

        except (KeyError, TypeError):
            raise KeyError(sub_path)


    # ----------------------------------------------------------------------------------------------------------------
    # target...

    # Copies from other at sub_path...

    def copy(self, other, sub_path=None):
        if sub_path is None:
            self.__dictionary = deepcopy(other.__dictionary)
            return

        nodes = re.findall('([^.:]+)([.:]*)', sub_path)

        try:
            self.__append(self.__dictionary, nodes, other.node(sub_path))

        except KeyError:
            raise KeyError(sub_path)


    # ----------------------------------------------------------------------------------------------------------------

    def __paths(self, container, prefix=None):
        # dict...
        if isinstance(container, dict):
            separator = '.'
            keys = container.keys()

        # list...
        elif isinstance(container, list):
            separator = ':'
            keys = range(len(container))

        # scalar...
        else:
            return [prefix]

        prefix = prefix + separator if prefix else ''

        paths = []

        for key in keys:
            path = prefix + str(key)
            value = container[key]

            # dict...
            if isinstance(value, dict):
                paths.extend(self.__paths(value, path))

            # list...
            elif isinstance(value, list):
                for i in range(len(value)):
                    if isinstance(value[i], (list, dict)):
                        paths.extend(self.__paths(value[i], path + ':' + str(i)))
                    else:
                        paths.append(path + ':' + str(i))

            # scalar...
            else:
                paths.append(path)

        return paths


    def __node(self, container, nodes):
        # key...
        if isinstance(container, list):
            try:
                key = int(nodes[0])

            except ValueError:
                raise KeyError(nodes[0])            # a non-integer key on an array is a KeyError

        else:
            key = nodes[0]

        # value...
        value = container[key]                      # raises TypeError if container is None

        # leaf...
        if len(nodes) == 1:
            return value

        # deep...
        return self.__node(value, nodes[1:])


    def __set_node(self, container, nodes, value):
        # key...
        if isinstance(container, list):
            try:
                key = int(nodes[0])

            except ValueError:
                raise KeyError(nodes[0])            # a non-integer key on an array is a KeyError

        else:
            key = nodes[0]

        # value...
        container[key] = value

        # leaf...
        if len(nodes) == 1:
            return value

        # deep...
        return self.__set_node(value, nodes[1:], value)


    def __append(self, container, nodes, value):
        # key...
        key = nodes[0][0]
        separator = nodes[0][1]

        # dict...
        if isinstance(container, dict):
            if key not in container:
                container[key] = [] if separator == ':' else OrderedDict()

        # list...
        elif isinstance(container, list):
            try:
                key = int(key)
            except ValueError:
                raise KeyError(key)

            while key >= len(container):
                container.append(None)

            if container[key] is None:
                container[key] = [] if separator == ':' else OrderedDict()

        # leaf...
        if len(nodes) == 1:
            container[key] = deepcopy(value)
            return

        # deep...
        self.__append(container[key], nodes[1:], value)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def dictionary(self):
        return self.node()


    @property
    def sort_paths(self):
        return self.__sort_paths


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        return self.__dictionary


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "PathDict:{sort_paths:%s, dictionary:%s}" % (self.sort_paths, self.node())



"""
Created on 19 Apr 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

A tool to remove outlier (excessively low) minimum values...
note that we sort the list in order to minimise the number of required tests.
"""

from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class MinList(object):
    """
    classdocs
   """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, length):
        """
        Constructor
        """
        self.__length = length                  # int
        self.__minimums = []                    # array of comparable


    # ----------------------------------------------------------------------------------------------------------------

    def append(self, value):
        mins_length = len(self.__minimums)

        # seed the list...
        if mins_length < self.__length:
            self.__minimums.append(value)
            self.__minimums.sort(reverse=True)
            return

        # ignore non-minimals...
        if value >= self.__minimums[0]:
            return

        # replace highest possible...
        for i in range(mins_length):
            if value < self.__minimums[i]:
                self.__minimums[i] = value
                self.__minimums.sort(reverse=True)
                return


    def max(self):
        if not self.__minimums:
            return None

        # return highest...
        return self.__minimums[0]


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        minimums = Str.collection(self.__minimums)

        return "MinList:{length:%s, minimums:%s}" % (self.__length, minimums)



"""
Created on 21 Jan 2021

@author: Jade Page (jade.page@southcoastscience.com)
"""

from scs_core.aws.data.deployment import Deployment
from scs_core.aws.greengrass.aws_group import AWSGroup


# --------------------------------------------------------------------------------------------------------------------

class AWSDeploymentReporter(object):

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, client):
        self.__client = client


    # ----------------------------------------------------------------------------------------------------------------

    def get_all_group_ids(self):
        next_token = None
        group_ids = []

        while True:
            response = self.__get_groups(next_token)

            if "Groups" not in response:
                break

            for item in response["Groups"]:
                group_ids.append(item.get("Id"))

            if "NextToken" not in response:
                break

            next_token = response["NextToken"]

        return group_ids


    def get_group_id(self, group_name):
        aws_group_info = AWSGroup(group_name, self.__client)
        datum = aws_group_info.get_group_info_from_name()

        return None if datum is None else datum.node("GroupID")


    def get_group_names(self, group_ids, matching=None, currency=None):
        return tuple(deployment.group_name for deployment in
                     self.get_deployments(group_ids, matching=matching, currency=currency))


    def get_deployments(self, group_ids, matching=None, currency=None):
        deployments = []

        for id in group_ids:
            if id is None:
                continue

            response = self.__client.list_deployments(GroupId=id)
            group_name = self.__get_group_name(id)

            if matching and matching not in group_name:
                continue

            latest_deployment = response["Deployments"][0] if response["Deployments"] else None
            deployment = Deployment.construct_from_aws(group_name, latest_deployment)

            if currency is None or not deployment.is_current(currency):
                deployments.append(deployment)

        return sorted(deployments)


    # ----------------------------------------------------------------------------------------------------------------

    def __get_groups(self, next_token):
        if next_token is None:
            return self.__client.list_groups()

        return self.__client.list_groups(NextToken=next_token)


    def __get_group_name(self, group_id):
        response = self.__client.get_group(GroupId=group_id)

        if "Name" not in response:
            return None

        return response["Name"]


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AWSDeploymentReporter:{client:%s}" % self.__client



"""
Created on 17 Apr 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import json
import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint

from scs_core.control.control_receipt import ControlReceipt

from scs_core.data.json import JSONify


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('DevCtrlAPI/DeviceControl',
                   'https://4fq7dy8f15.execute-api.us-west-2.amazonaws.com/default/DeviceControl')


# --------------------------------------------------------------------------------------------------------------------

class DeviceControlClient(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def interact(self, token, device_tag, cmd_tokens):
        payload = {
            'device-tag': device_tag,
            'message': [str(token) for token in cmd_tokens]
        }

        response = requests.post(Endpoint.url(), headers=self._token_headers(token), data=JSONify.dumps(payload))
        self._check_response(response)

        return ControlReceipt.construct_from_jdict(json.loads(response.json()))



"""
Created on 4 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

AQCSV: Measurement performance characteristics

NB: initialisation is performed at the foot of this class

example:
{"code": 2, "abbreviation": "XC", "definition": "Critical Value",
"description": "Threshold above which a measurement result is unlikely to result from a true value of zero"}

https://www.airnow.gov/
"""

import os

from collections import OrderedDict

from scs_core.csv.csv_archive import CSVArchive
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class MPC(CSVArchive, JSONable):
    """
    classdocs
    """

    _retrieved = {}

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def archive_location(cls):
        return os.path.join(os.path.dirname(os.path.realpath(__file__)), 'archive', 'mpcs.csv')


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        code = jdict.get('code')
        abbreviation = jdict.get('abbreviation')
        definition = jdict.get('definition')
        description = jdict.get('description')

        return MPC(code, abbreviation, definition, description)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, code, abbreviation, definition, description):
        """
        Constructor
        """
        self.__code = int(code)                                 # int
        self.__abbreviation = abbreviation                      # string
        self.__definition = definition                          # string
        self.__description = description                        # string


    def __eq__(self, other):
        try:
            return self.code == other.code and self.abbreviation == other.abbreviation and \
                   self.definition == other.definition and self.description == other.description

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['code'] = self.code
        jdict['abbreviation'] = self.abbreviation
        jdict['definition'] = self.definition
        jdict['description'] = self.description

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def pk(self):
        return self.code


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def code(self):
        return self.__code


    @property
    def abbreviation(self):
        return self.__abbreviation


    @property
    def definition(self):
        return self.__definition


    @property
    def description(self):
        return self.__description


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MPC:{code:%d, abbreviation:%s, definition:%s, description:%s}" % \
               (self.code, self.abbreviation, self.definition, self.description)


# --------------------------------------------------------------------------------------------------------------------
# initialisation...

MPC.retrieve()



"""
Created on 14 Jun 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

document example:
{"scs-opc-109": ["somebody@somewhere.com", "somebody@somewhere-else.com"], ...}
"""

from collections import OrderedDict

from scs_core.data.json import PersistentJSONable
from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class DeviceMonitorEmailList(PersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "device_monitor_email_list.json"

    @classmethod
    def persistence_location(cls):
        return cls.aws_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if jdict is None:
            return None

        device_dict = {device_tag: set(recipients) for device_tag, recipients in jdict.items()}

        return cls(device_dict)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device_dict):
        """
        Constructor
        """
        super().__init__()

        self.__device_dict = device_dict                            # dict of device-tag: set of emails


    def __len__(self):
        return len(self.__device_dict)


    # ----------------------------------------------------------------------------------------------------------------

    def add(self, device_tag, email_address):
        if device_tag not in self.__device_dict:
            self.__device_dict[device_tag] = set()

        self.__device_dict[device_tag].add(email_address)

        return DeviceMonitorEmailList({device_tag: sorted(self.__device_dict[device_tag])})


    def discard(self, device_tag, email_address):
        device_dict = {}

        # all devices...
        if device_tag is None:
            for device_tag in self.__device_dict:
                self.__device_dict[device_tag].discard(email_address)
                device_dict[device_tag] = self.__device_dict[device_tag]

        # specific device...
        elif device_tag not in self.__device_dict:
            device_dict = {}

        else:
            self.__device_dict[device_tag].discard(email_address)
            device_dict = {device_tag: sorted(self.__device_dict[device_tag])}

        return DeviceMonitorEmailList(device_dict)


    def filter(self, email_address=None, device_tag=None, exact=False):
        device_dict = {}

        for key in self.__device_dict.keys():
            if not self.__matches_tag(key, device_tag, exact):
                continue

            if not self.__matches_email(self.__device_dict[key], email_address, exact):
                continue

            device_dict[key] = self.__device_dict[key]

        return DeviceMonitorEmailList(device_dict)


    def filter_devices(self, device_tags):
        device_dict = {key: self.__device_dict[key] for key in self.__device_dict.keys() & device_tags}

        return DeviceMonitorEmailList(device_dict)


    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def __matches_tag(key, device_tag, exact):
        if device_tag is None:
            return True

        return (exact and device_tag == key) or (not exact and device_tag in key)


    @staticmethod
    def __matches_email(email_list, email_address, exact):
        if email_address is None:
            return True

        for email in email_list:
            if (exact and email_address == email) or (not exact and email_address in email):
                return True

        return False


    # ----------------------------------------------------------------------------------------------------------------

    def subset(self, email_address):
        device_dict = {device_tag: recipients for device_tag, recipients in self.__device_dict.items()
                       if email_address in recipients}

        return DeviceMonitorEmailList(device_dict)


    def recipients(self, device_tag):
        try:
            return self.__device_dict[device_tag]
        except KeyError:
            return None


    # ----------------------------------------------------------------------------------------------------------------

    def device(self, device_tag):
        try:
            return self.__device_dict[device_tag]

        except KeyError:
            return None


    @property
    def device_dict(self):
        device_dict = OrderedDict()
        for device_tag in sorted(self.__device_dict):
            if self.__device_dict[device_tag]:
                device_dict[device_tag] = sorted(self.__device_dict[device_tag], key=lambda email: email.lower())

        return device_dict


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        return self.device_dict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "DeviceMonitorEmailList:{device_dict:%s}" %  Str.collection(self.device_dict)



"""
Created on 3 May 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://stackoverflow.com/questions/6677424/how-do-i-import-variable-packages-in-python-like-using-variable-variables-i

example JSON:
{"scs_core": {"repo": "scs_core", "version": "1.0.5"}, "scs_dev": {"repo": "scs_dev", "version": "1.0.4"},
"scs_dfe": {"repo": "scs_dfe_eng", "version": "1.0.2"}, "scs_host": {"repo": "scs_host_cpc", "version": "1.0.2"},
"scs_inference": {"repo": "scs_inference", "version": null}, "scs_mfr": {"repo": "scs_mfr", "version": "1.0.2"},
"scs_ndir": {"repo": "scs_ndir", "version": "0.9.1"}, "scs_psu": {"repo": "scs_psu", "version": "1.0.5"}}
"""

import os

from collections import OrderedDict
from importlib import import_module

from scs_core.data.json import JSONable, MultiPersistentJSONable
from scs_core.data.str import Str

from scs_core.estate.git_pull import GitPull
from scs_core.estate.software_version import SoftwareVersion

from scs_core.sys.filesystem import Filesystem


# --------------------------------------------------------------------------------------------------------------------

class PackageVersion(MultiPersistentJSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    __FILENAME = "package_version.json"

    @classmethod
    def persistence_location(cls, name):
        filename = cls.__FILENAME if name is None else '_'.join((name, cls.__FILENAME))

        return cls.conf_dir(), filename


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_installation(cls, package, repository):
        try:
            module = import_module(package)
        except ModuleNotFoundError:
            return cls(package, repository, None)

        try:
            # noinspection PyUnresolvedReferences
            version = SoftwareVersion.construct_from_jdict(module.__version__)
        except AttributeError:
            version = None

        return cls(package, repository, version)


    @classmethod
    def construct_from_jdict(cls, jdict, name=None, skeleton=False):
        if not jdict:
            return None

        repository = jdict.get('repo')
        version = SoftwareVersion.construct_from_jdict(jdict.get('version'))

        return cls(name, repository, version)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, name, repository, version):
        """
        Constructor
        """
        super().__init__(name)

        self.__repository = repository              # string
        self.__version = version                    # SoftwareVersion or None


    def __eq__(self, other):
        try:
            return self.repository == other.repository and self.version == other.version

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def repository(self):
        return self.__repository


    @property
    def version(self):
        return self.__version


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['repo'] = self.repository
        jdict['version'] = self.version

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "PackageVersion:{name:%s, repository:%s, version:%s}" %  (self.name, self.repository, self.version)


# --------------------------------------------------------------------------------------------------------------------

class PackageVersions(JSONable):
    """
    classdocs
    """

    __GREENGRASS_PACKAGE = 'scs_greengrass'

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_installation(cls, root, manager):
        versions = {}

        # local...
        for repository in GitPull.dirs(root):
            contents = Filesystem.ls(os.path.join(root, repository, 'src'))

            for content in contents:
                package = content.name
                versions[package] = PackageVersion.construct_from_installation(package, repository)

        # greengrass...
        greengrass_version = PackageVersion.load(manager, name=cls.__GREENGRASS_PACKAGE)

        if greengrass_version is not None:
            versions[cls.__GREENGRASS_PACKAGE] = greengrass_version

        return cls(versions)


    @classmethod
    def construct_from_jdict(cls, jdict):
        if jdict is None:
            return None

        versions = {}

        for package, version in jdict.items():
            versions[package] = PackageVersion.construct_from_jdict(version, name=package)

        return cls(versions)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, versions):
        """
        Constructor
        """
        self.__versions = versions                  # dict of package: PackageVersion


    def __len__(self):
        return len(self.versions)


    def __eq__(self, other):
        try:
            if len(self) != len(other):
                return False

            for package in self.__versions.keys():
                if self.__versions[package] != other.__versions[package]:
                    return False

            return True

        except (KeyError, TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def version(self, package):
        try:
            return self.__versions[package]

        except KeyError:
            return None


    @property
    def versions(self):
        return {package: self.__versions[package] for package in sorted(self.__versions.keys())}


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        return self.versions


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "PackageVersions:{versions:%s}" %  Str.collection(self.versions)



"""
Created on 9 Dec 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

The A4CalibratedDatum is designed to provide a model training data set that encapsulates the calibration of the
electrochemical sensor - the fields we_v_cal, ae_v_zero_cal and cal_x_v have no meaning beyond this.
cal_x_v is only relevant to sensors with NO2 cross-sensitivity.
"""

from collections import OrderedDict

from scs_core.data.datum import Datum

from scs_core.gas.a4.a4_calib import A4Calib
from scs_core.gas.a4.a4_datum import A4Datum


# --------------------------------------------------------------------------------------------------------------------

class A4Calibrator(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, calib: A4Calib):
        """
        Constructor
        """
        self.__calib = calib                                                # A4Calib

        self.__we_elc_v = calib.we_elc_mv / 1000.0
        self.__ae_elc_v = calib.ae_elc_mv / 1000.0

        self.__we_sens_v = calib.we_sens_mv / 1000.0

        self.__we_no2_x_sens_v = None if calib.we_no2_x_sens_mv is None else calib.we_no2_x_sens_mv / 1000.0


    # ----------------------------------------------------------------------------------------------------------------

    def calibrate(self, datum, no2_cnc=None):
        # zero offset...
        we_v_zero_cal = datum.we_v - self.__we_elc_v
        ae_v_zero_cal = datum.ae_v - self.__ae_elc_v

        # gain...
        we_v_cal = we_v_zero_cal / self.__we_sens_v

        # cross sensitivity...
        cal_x_v = None if no2_cnc is None else no2_cnc * self.__we_no2_x_sens_v

        return A4CalibratedDatum(datum.we_v, datum.ae_v, datum.we_c, datum.cnc, we_v_cal, ae_v_zero_cal, cal_x_v)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "A4Calibrator:{calib:%s}" % self.__calib


# --------------------------------------------------------------------------------------------------------------------

class A4CalibratedDatum(A4Datum):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, we_v, ae_v, we_c, cnc, we_v_cal, ae_v_cal, cal_x_v):
        """
        Constructor
        """
        super().__init__(we_v, ae_v, we_c, cnc)

        self.__we_v_cal = Datum.float(we_v_cal, 6)              # calibrated WE voltage
        self.__ae_v_cal = Datum.float(ae_v_cal, 6)              # calibrated AE voltage

        self.__cal_x_v = Datum.float(cal_x_v, 9)                # calibrated cross-sensitivity voltage


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['weV'] = self.we_v
        jdict['aeV'] = self.ae_v

        jdict['weC'] = self.we_c                            # may be None
        jdict['cnc'] = self.cnc                             # may be None

        jdict['weVCal'] = self.we_v_cal
        jdict['aeVCal'] = self.ae_v_cal

        if self.cal_x_v is not None:
            jdict['calXV'] = self.cal_x_v

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def we_v_cal(self):
        return self.__we_v_cal


    @property
    def ae_v_cal(self):
        return self.__ae_v_cal


    @property
    def cal_x_v(self):
        return self.__cal_x_v


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "A4CalibratedDatum(v1):{we_v:%s, ae_v:%s, we_c:%s, cnc:%s, we_v_cal:%s, ae_v_cal:%s, cal_x_v:%s}" % \
               (self.we_v, self.ae_v, self.we_c, self.cnc, self.we_v_cal, self.ae_v_cal, self.cal_x_v)



"""
Created on 13 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

MappingTask example:
{"org": "south-coast-science-demo", "group": "brighton", "loc": 1, "topic": "particulates", "device": "praxis-000401",
"parameters": ["pm1", "pm2p5", "pm10"], "duration": 1, "checkpoint": "**:/01:00", "agency-code": "AAA",
"site-code": "123MM123456789", "pocs": {}, "upload-start": "2019-02-01T00:00:00Z", "upload-end": null}
"""

from ast import literal_eval

from collections import OrderedDict

from scs_core.aqcsv.connector.datum_mapping import DatumMapping

from scs_core.aqcsv.data.aqcsv_datetime import AQCSVDatetime
from scs_core.aqcsv.data.aqcsv_site import AQCSVSite

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable, PersistentJSONable
from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class AirNowMappingTaskList(PersistentJSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    __FILENAME =    "airnow_mapping_tasks.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return AirNowMappingTaskList({}) if skeleton else None

        tasks = {literal_eval(key): MappingTask.construct_from_jdict(task) for key, task in jdict.get('tasks').items()}

        return AirNowMappingTaskList(tasks)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tasks):
        """
        Constructor
        """
        super().__init__()

        self.__tasks = tasks                                        # dictionary of task.pk: task


    # ----------------------------------------------------------------------------------------------------------------

    def items(self):
        return self.__tasks.values()


    def item(self, pk):
        if pk not in self.__tasks:
            return None

        return self.__tasks[pk]


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['tasks'] = {str(key): task for key, task in self.__tasks.items()}

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def insert(self, task):
        self.__tasks[task.pk] = task

        # sort...
        tasks = OrderedDict()

        for pk in sorted(self.__tasks.keys()):
            tasks[pk] = self.__tasks[pk]

        self.__tasks = tasks


    def remove(self, pk):
        try:
            del self.__tasks[pk]

        except KeyError:
            pass


    def set_upload_start(self, pk, upload_start):
        self.__tasks[pk].upload_start = upload_start


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AirNowMappingTaskList:{tasks:%s}" % Str.collection(self.__tasks)


# --------------------------------------------------------------------------------------------------------------------

class MappingTask(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        org = jdict.get('org')
        group = jdict.get('group')
        loc = jdict.get('loc')
        topic = jdict.get('topic')

        device = jdict.get('device')
        parameters = jdict.get('parameters')
        duration = jdict.get('duration')
        checkpoint = jdict.get('checkpoint')

        agency_code = jdict.get('agency-code')
        site_code = jdict.get('site-code')
        pocs = jdict.get('pocs')

        upload_start = LocalizedDatetime.construct_from_jdict(jdict.get('upload-start'))
        upload_end = LocalizedDatetime.construct_from_jdict(jdict.get('upload-end'))

        return MappingTask(org, group, loc, topic, device, parameters, duration, checkpoint,
                           agency_code, site_code, pocs, upload_start, upload_end)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, org, group, loc, topic, device, parameters, duration, checkpoint,
                 agency_code, site_code, pocs, upload_start, upload_end):
        """
        Constructor
        """
        self.__org = org                                    # string
        self.__group = group                                # string
        self.__loc = int(loc)                               # int
        self.__topic = topic                                # string

        self.__device = device                              # string
        self.__parameters = tuple(parameters)               # tuple of string
        self.__duration = int(duration)                     # int                       minutes
        self.__checkpoint = checkpoint                      # string

        self.__agency_code = agency_code                    # string
        self.__site_code = site_code                        # string
        self.__pocs = pocs                                  # dictionary of parameter: index

        self.__upload_start = upload_start                  # LocalizedDatetime
        self.__upload_end = upload_end                      # LocalizedDatetime


    def __eq__(self, other):
        try:
            return self.org == other.org and \
                   self.group == other.group and \
                   self.loc == other.loc and \
                   self.topic == other.topic and \
                   self.device == other.device and \
                   self.parameters == other.parameters and \
                   self.duration == other.duration and \
                   self.checkpoint == other.checkpoint and \
                   self.agency_code == other.agency_code and \
                   self.site_code == other.site_code and \
                   self.pocs == other.pocs and \
                   self.upload_start == other.upload_start and \
                   self.upload_end == other.upload_end

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['org'] = self.org
        jdict['group'] = self.group
        jdict['loc'] = self.loc
        jdict['topic'] = self.topic

        jdict['device'] = self.device
        jdict['parameters'] = self.parameters
        jdict['duration'] = self.duration
        jdict['checkpoint'] = self.checkpoint

        jdict['agency-code'] = self.agency_code
        jdict['site-code'] = self.site_code
        jdict['pocs'] = self.pocs

        jdict['upload-start'] = self.upload_start
        jdict['upload-end'] = self.upload_end

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def environment_path(self):
        return '/'.join((self.org, self.group, 'loc', str(self.loc), self.topic))


    def status_path(self):
        return '/'.join((self.org, self.group, 'device', self.device, 'status'))


    def mappings(self):
        return tuple(DatumMapping(self.topic, species, self.site_code) for species in self.parameters)


    def file_prefix(self):
        dt = AQCSVDatetime(LocalizedDatetime.now().utc().datetime)
        site = AQCSVSite.construct_from_code(self.site_code)

        return dt.filename_prefix() + '_' + str(site.country_code)          # goal is YYYYMMDDhhmm_CCC.AAAAAAAAAA


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def pk(self):
        return self.org, self.group, self.loc, self.topic


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def org(self):
        return self.__org


    @property
    def group(self):
        return self.__group


    @property
    def loc(self):
        return self.__loc


    @property
    def topic(self):
        return self.__topic


    @property
    def device(self):
        return self.__device


    @property
    def parameters(self):
        return self.__parameters


    @property
    def checkpoint(self):
        return self.__checkpoint


    @property
    def duration(self):
        return self.__duration


    @property
    def agency_code(self):
        return self.__agency_code


    @property
    def site_code(self):
        return self.__site_code


    @property
    def pocs(self):
        return self.__pocs


    @property
    def upload_start(self):
        return self.__upload_start


    @property
    def upload_end(self):
        return self.__upload_end


    @upload_end.setter
    def upload_end(self, upload_end):
        self.__upload_end = upload_end


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MappingTask:{org:%s, group:%s, loc:%s, topic:%s, device:%s, parameters:%s, duration:%s, " \
               "checkpoint:%s, agency_code:%s, site_code:%s, pocs:%s, upload_start:%s, upload_end:%s}" % \
               (self.org, self.group, self.loc, self.topic, self.device, self.parameters, self.duration,
                self.checkpoint, self.agency_code, self.site_code, self.pocs, self.upload_start, self.upload_end)



"""
Created on 18 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Alphasense Application Note AAN 803-02
AAN 803-02 070916_DRAFT03.doc
"""

from collections import OrderedDict

from scs_core.data.datum import Datum
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class A4Datum(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        we_v = jdict.get('weV')
        ae_v = jdict.get('aeV')

        we_c = jdict.get('weC')
        cnc = jdict.get('cnc')

        return cls(we_v, ae_v, we_c, cnc)


    @classmethod
    def construct(cls, calib, baseline, tc, temp, we_v, ae_v, no2_cnc=None):
        if calib is None or tc is None:
            return A4Datum(we_v, ae_v)

        # print("A4Datum: calib:%s baseline:%s tc:%s temp:%f we_v:%f ae_v:%f x_sens_sample:%s" %
        #       (calib, baseline, tc, temp, we_v, ae_v, x_sens_sample), file=sys.stderr)

        # weC...
        we_c = cls.__we_c(calib, tc, temp, we_v, ae_v)

        if we_c is None:
            return A4Datum(we_v, ae_v)

        if no2_cnc:
            we_c -= cls.__reverse_we_c(calib.we_no2_x_sens_mv, no2_cnc)

        # cnc...
        cnc = cls.__cnc(calib.we_sens_mv, we_c)

        baselined_cnc = cnc + baseline.offset                # a positive offset causes the value to be raised

        return cls(we_v, ae_v, we_c, baselined_cnc)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def __we_c(cls, calib, tc, temp, we_v, ae_v):
        """
        Compute weC from sensor temperature compensation of weV, aeV
        """
        we_t = we_v - (calib.we_elc_mv / 1000.0)        # remove electronic we zero
        ae_t = ae_v - (calib.ae_elc_mv / 1000.0)        # remove electronic ae zero

        we_c = tc.correct(calib, temp, we_t, ae_t)

        # print("A4Datum__we_c: we_t:%f ae_t:%f we_c:%s" % (we_t, ae_t, we_c), file=sys.stderr)

        return we_c


    @classmethod
    def __cnc(cls, sens_mv, we_c):
        """
        Compute cnc from weC (using primary sensitivity)
        """
        if we_c is None:
            return None

        cnc = we_c / (sens_mv / 1000.0)

        # print("A4Datum__cnc: we_c:%s cnc:%f" % (we_c, cnc), file=sys.stderr)

        return cnc


    @classmethod
    def __reverse_we_c(cls, sens_mv, cnc):      # TODO: don't use concentration - factor out baseline?
        """
        Compute weC from cnc (using cross-sensitivity)
        """
        we_c = (cnc * sens_mv) / 1000.0

        # print("__reverse_we_c: sens_mv:%s cnc:%f we_c:%s" % (sens_mv, cnc, we_c), file=sys.stderr)

        return we_c


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, we_v, ae_v, we_c=None, cnc=None):
        """
        Constructor
        """
        self.__we_v = Datum.float(we_v, 5)        # uncorrected working electrode voltage       Volts
        self.__ae_v = Datum.float(ae_v, 5)        # uncorrected auxiliary electrode voltage     Volts

        self.__we_c = Datum.float(we_c, 5)        # corrected working electrode voltage         Volts
        self.__cnc = Datum.float(cnc, 1)          # gas concentration                           ppb


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['weV'] = self.we_v
        jdict['aeV'] = self.ae_v

        jdict['weC'] = self.we_c                 # may be None
        jdict['cnc'] = self.cnc                  # may be None

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def we_v(self):
        return self.__we_v


    @property
    def ae_v(self):
        return self.__ae_v


    @property
    def we_c(self):
        return self.__we_c


    @property
    def cnc(self):
        return self.__cnc


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "A4Datum:{we_v:%s, ae_v:%s, we_c:%s, cnc:%s}" % \
               (self.we_v, self.ae_v, self.we_c, self.cnc)



"""
Created on 25 Feb 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

an empty catalogue of particulate exegesis models, to be implemented elsewhere
"""


# --------------------------------------------------------------------------------------------------------------------

class ExegeteCatalogue(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def model_names():
        return tuple()


    @staticmethod
    def load(_name, _host):
        raise NotImplementedError


    @staticmethod
    def standard(_name):
        raise NotImplementedError



"""
Created on 14 Jul 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

A portable, non-blocking Python line reader. Yes, such a thing exists. And this is it...
"""

import os

from multiprocessing import Process, Queue


# --------------------------------------------------------------------------------------------------------------------

class LineReader(object):
    """
    classdocs
    """
    def __init__(self, fileno):
        """
        Constructor
        """
        self.__fileno = fileno
        self.__queue = Queue()


    # ----------------------------------------------------------------------------------------------------------------

    def start(self):
        proc = Process(target=self.run)
        proc.start()

        return proc


    def run(self):
        file = os.fdopen(self.__fileno)

        try:
            for line in file:
                self.__queue.put(str(line).strip())

        except (ConnectionError, KeyboardInterrupt, SystemExit):
            pass

        self.__queue.put(None)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def lines(self):
        while True:
            if self.__queue.empty():
                line = None

            else:
                line = self.__queue.get()

                if line is None:
                    return

            yield line


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "LineReader:{fileno:%s, queue:%s}" % (self.__fileno, self.__queue)



"""
Created on 22 Feb 2021

@author: Jade Page (jade.page@southcoastscience.com)
"""

import json
import logging
import requests
import time

from collections import OrderedDict
from AWSIoTPythonSDK.exception import operationError, operationTimeoutException

from scs_core.aws.client.client_auth import ClientAuth
from scs_core.aws.client.mqtt_client import MQTTClient, MQTTSubscriber

from scs_core.aws.manager.byline.byline import TopicBylineGroup
from scs_core.aws.manager.dynamo_manager import DynamoManager

from scs_core.control.control_datum import ControlDatum
from scs_core.control.control_handler import ControlHandler

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.publication import Publication

from scs_core.estate.configuration import Configuration


# ----------------------------------------------------------------------------------------------------------------

class MQTTDevicePoller(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    __DEVICE_TOPICS_URL = 'https://xy1eszuu23.execute-api.us-west-2.amazonaws.com/dev/device-topics'

    __BUCKET = "scs-persistence"
    __KEY = "conf/mqtt_peers.json"
    __TABLE = "device_configuration"

    __TIMEOUT = 30          # seconds

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, host, s3_manager, dynamo_client=None, dynamo_resource=None):
        """
        Constructor
        """
        self.__host = host                                          # PersistenceManager
        self.__s3_manager = s3_manager
        self.__dynamo_manager = DynamoManager(dynamo_client, dynamo_resource) if dynamo_client and dynamo_resource \
            else None

        self.__logger = logging.getLogger()


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def all_devices(cls):
        req = requests.get(cls.__DEVICE_TOPICS_URL)
        status = req.status_code
        if status != 200:
            return None

        res = req.content
        jdict = json.loads(res)

        group = TopicBylineGroup.construct_from_jdict(jdict, excluded="/control")

        return group.devices


    def known_devices(self, with_info=False):
        # TODO: replace the below with load(..) method
        res, _ = self.__s3_manager.retrieve_from_bucket(self.__BUCKET, self.__KEY)
        json_data = json.loads(res)
        data = json_data["peers"]

        known_dev = []

        if not with_info:
            for item in data.items():
                tag = item[1]["tag"]
                known_dev.append(tag)
        else:
            for item in data.items():
                data = item[1]
                known_dev.append(data)

        return known_dev


    def missing_devices(self):
        missing_dev = set(self.all_devices()) - set(self.known_devices())

        return sorted(missing_dev)


    # ----------------------------------------------------------------------------------------------------------------

    def update_configs(self):
        jdict = OrderedDict()

        for device in self.known_devices(with_info=True):
            d_tag = device["tag"]
            d_ss = device["shared-secret"]
            d_topic = device["topic"]
            res = self.send_mqtt(d_tag, d_ss, d_topic, "?")

            jdict[d_tag] = res
            print("Device:%s :%s" % (d_tag, res))
            self.__logger.info("Device:%s :%s" % (d_tag, res))

            if res != "Timeout":
                if "configuration" in res[0]:
                    self.get_configuration(d_tag, d_ss, d_topic)


    def get_configuration(self, device_tag, shared_secret, topic):
        tokens = ["configuration"]
        print(type(tokens))
        res = self.send_mqtt(device_tag, shared_secret, topic, tokens)
        self.save_changes(device_tag, res)


    def send_mqtt(self, device_tag, d_ss, d_topic, token):
        # tag...
        host_tag = self.__host.name()

        # ClientAuth...
        auth = ClientAuth.load(self.__host)

        if auth is None:
            # log no auth
            exit(1)

        # responder...
        handler = ControlHandler(host_tag, device_tag)

        subscriber = MQTTSubscriber(d_topic, handler.handle)
        client = MQTTClient(subscriber)

        # go
        client.connect(auth)

        # datum...
        now = LocalizedDatetime.now().utc()
        datum = ControlDatum.construct(host_tag, device_tag, now, token, self.__TIMEOUT, d_ss)

        publication = Publication(d_topic, datum)

        handler.set_outgoing(publication)

        try:
            client.publish(publication)

        except (OSError, operationError, operationTimeoutException) as ex:
            self.__logger.error(repr(ex))

        timeout = time.time() + self.__TIMEOUT

        while True:
            if handler.receipt:
                if not handler.receipt.is_valid(d_ss):
                    client.disconnect()
                    raise ValueError("invalid digest: %s" % handler.receipt)

                if handler.receipt.command.stderr:
                    client.disconnect()
                    return handler.receipt.command.stderr

                if handler.receipt.command.stdout:
                    client.disconnect()
                    return handler.receipt.command.stdout

            if time.time() > timeout:  # was cmd.interactive and ...
                client.disconnect()
                return "Timeout"

            time.sleep(0.1)


    def save_changes(self, device_tag, response):
        # see if item already in DB

        current = self.__dynamo_manager.items(self.__TABLE, "device_tag", device_tag)
        if current:
            loaded = current[0]['data'][0]
            res_str = response[0]
            loaded_conf = Configuration.construct_from_jstr(loaded)
            this_conf = Configuration.construct_from_jstr(res_str)
            if loaded_conf == this_conf:
                # config remains the same, delete original compo key & replace to avoid clutter
                key = {
                    'device_tag': current[0]['device_tag'],
                    'datetime': current[0]['datetime']
                }

                self.__dynamo_manager.delete_item(self.__TABLE, key)

        # add to db
        item = {
            'device_tag': device_tag,
            'datetime': LocalizedDatetime.now().as_iso8601(),
            'data': response
        }

        self.__dynamo_manager.add(self.__TABLE, item)





"""
Created on 27 Jan 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

    VERSION = 1.3

example document:
{
    "rec": "2024-04-04T09:28:16Z",
    "tag": "scs-be2-3",
    "ver": 1.4,
    "val": {
        "hostname": "scs-bbe-003",
        "platform": {
            "os": "10.13",
            "kernel": "6.1.77-bone30"
        },
        "packs": {
            "scs_comms": {
                "repo": "scs_comms_ge910",
                "version": null
            },
            "scs_core": {
                "repo": "scs_core",
                "version": "3.11.2"
            },
            "scs_dev": {
                "repo": "scs_dev",
                "version": "3.4.5"
            },
            "scs_dfe": {
                "repo": "scs_dfe_eng",
                "version": "3.2.2"
            },
            "scs_exegesis": {
                "repo": "scs_exegesis",
                "version": null
            },
            "scs_greengrass": {
                "repo": "scs_greengrass",
                "version": "2.5.0"
            },
            "scs_host": {
                "repo": "scs_host_bbe_southern",
                "version": "3.5.2"
            },
            "scs_inference": {
                "repo": "scs_inference",
                "version": null
            },
            "scs_mfr": {
                "repo": "scs_mfr",
                "version": "3.8.11"
            },
            "scs_ndir": {
                "repo": "scs_ndir",
                "version": null
            },
            "scs_psu": {
                "repo": "scs_psu",
                "version": "2.6.2"
            }
        },
        "afe-baseline": {
            "sn1": {
                "calibrated-on": "2023-12-07T12:43:55Z",
                "offset": 0
            },
            "sn2": {
                "calibrated-on": "2023-12-07T12:43:55Z",
                "offset": 0
            },
            "sn3": {
                "calibrated-on": "2023-12-07T12:43:55Z",
                "offset": 0
            },
            "sn4": {
                "calibrated-on": "2023-12-07T12:43:55Z",
                "offset": 0
            }
        },
        "afe-id": {
            "serial_number": "26-000345",
            "type": "810-0023-01",
            "calibrated_on": "2020-11-18",
            "sn1": {
                "serial_number": "212632052",
                "sensor_type": "NO2A43F"
            },
            "sn2": {
                "serial_number": "214250436",
                "sensor_type": "OXA431"
            },
            "sn3": {
                "serial_number": "130631043",
                "sensor_type": "NO A4"
            },
            "sn4": {
                "serial_number": "134200204",
                "sensor_type": "SO2A4"
            }
        },
        "aws-group-config": {
            "group-name": "scs-bbe-003-group",
            "time-initiated": "2024-03-07T12:45:28Z",
            "unix-group": 987,
            "ml": "uE.1"
        },
        "aws-project": {
            "location-path": "south-coast-science-dev/development/loc/1",
            "device-path": "south-coast-science-dev/development/device"
        },
        "data-log": {
            "path": "/srv/removable_data_storage",
            "is-available": true,
            "on-root": false,
            "used": 6
        },
        "display-conf": null,
        "vcal-baseline": null,
        "gas-baseline": null,
        "gas-model-conf": null,
        "gps-conf": {
            "model": "SAM8Q",
            "sample-interval": 10,
            "tally": 60,
            "report-file": "/tmp/southcoastscience/gps_report.json",
            "debug": false
        },
        "interface-conf": {
            "model": "DFE"
        },
        "mpl115a2-calib": null,
        "opc-conf": {
            "model": "N3",
            "sample-period": 10,
            "restart-on-zeroes": true,
            "power-saving": false
        },
        "opc-version": {
            "serial": "177050912",
            "firmware": "OPC-N3 Iss1.1 FirmwareVer=1.17a...........................BS"
        },
        "opc-errors": 8,
        "pmx-model-conf": {
            "uds-path": "pipes/lambda-pmx-model.uds",
            "model-interface": "s2",
            "model-map": "uE.1"
        },
        "pressure-conf": null,
        "psu-conf": {
            "model": "OsloV1",
            "batt-model": null,
            "ignore-threshold": false,
            "reporting-interval": 10,
            "report-file": "/tmp/southcoastscience/psu_status_report.json"
        },
        "psu-version": {
            "id": "South Coast Science PSU Oslo",
            "tag": "2.2.5"
        },
        "scd30-baseline": null,
        "scd30-conf": null,
        "schedule": {
            "scs-climate": {
                "interval": 60.0,
                "tally": 1
            },
            "scs-gases": {
                "interval": 10.0,
                "tally": 1
            },
            "scs-particulates": {
                "interval": 10.0,
                "tally": 1
            },
            "scs-status": {
                "interval": 60.0,
                "tally": 1
            }
        },
        "sht-conf": {
            "int": "0x45",
            "ext": "0x45"
        },
        "networks": null,
        "modem": {
            "id": "992c3ac6da0b68d58005d20ea5e957d409001e42",
            "imei": "860425041573914",
            "mfr": "QUALCOMM INCORPORATED",
            "rev": "EC25ECGAR06A05M1G"
        },
        "sim": {
            "imsi": "234301951432537",
            "iccid": "8944303382697124823",
            "operator-code": "23430",
            "operator-name": "EE"
        },
        "system-id": {
            "set-on": "2024-01-09T16:02:09Z",
            "vendor-id": "SCS",
            "model-id": "BE2",
            "model": "Alpha BB Eng",
            "config": "V2",
            "system-sn": 3
        },
        "timezone-conf": {
            "set-on": "2017-08-15T12:50:05Z",
            "name": "Europe/London"
        }
    }
}
"""

import json
import socket

from collections import OrderedDict

from scs_core.aws.config.project import Project
from scs_core.aws.greengrass.aws_group_configuration import AWSGroupConfiguration

from scs_core.climate.mpl115a2_calib import MPL115A2Calib
from scs_core.climate.pressure_conf import PressureConf
from scs_core.climate.sht_conf import SHTConf

from scs_core.csv.csv_logger_conf import CSVLoggerConf

from scs_core.data.json import JSONable

from scs_core.display.display_conf import DisplayConf

from scs_core.estate.package_version import PackageVersions

from scs_core.gas.afe_baseline import AFEBaseline
from scs_core.gas.afe_id import AFEId
from scs_core.gas.scd30.scd30_conf import SCD30Conf
from scs_core.gas.scd30.scd30_baseline import SCD30Baseline

from scs_core.gps.gps_conf import GPSConf

from scs_core.interface.interface_conf import InterfaceConf

from scs_core.location.timezone_conf import TimezoneConf

from scs_core.model.gas.gas_baseline import GasBaseline
from scs_core.model.gas.gas_model_conf import GasModelConf
from scs_core.model.gas.vcal_baseline import VCalBaseline

from scs_core.model.pmx.pmx_model_conf import PMxModelConf

from scs_core.particulate.opc_conf import OPCConf
from scs_core.particulate.opc_error_log import OPCErrorSummary
from scs_core.particulate.opc_version import OPCVersion

from scs_core.psu.psu_conf import PSUConf
from scs_core.psu.psu_version import PSUVersion

from scs_core.sync.schedule import Schedule

from scs_core.sys.filesystem import FilesystemReport
from scs_core.sys.modem import Modem, SIM
from scs_core.sys.network import Networks
from scs_core.sys.platform import PlatformSummary
from scs_core.sys.system_id import SystemID


# --------------------------------------------------------------------------------------------------------------------

class Configuration(JSONable):
    """
    classdocs
    """

    VERSION = 1.4

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jstr(cls, jstr):
        try:
            return cls.construct_from_jdict(json.loads(jstr))
        except json.decoder.JSONDecodeError:
            return None


    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            if skeleton:
                return cls(None, None, None, None, None,
                           None, None, None, None, None,
                           None, None, None, None, None,
                           None, None, None, None, None,
                           None, None, None, None, None,
                           None, None, None, None, None,
                           None)
            else:
                return None

        hostname = jdict.get('hostname')

        if 'platform' in jdict:                 # current
            node = jdict.get('platform')
        else:                                   # deprecated
            os_node = jdict.get('os')
            node = {'kernel': os_node.get('rel')} if os_node else None

        platform = PlatformSummary.construct_from_jdict(node)

        packs = PackageVersions.construct_from_jdict(jdict.get('packs'))

        afe_baseline = AFEBaseline.construct_from_jdict(jdict.get('afe-baseline'))
        afe_id = AFEId.construct_from_jdict(jdict.get('afe-id'))
        aws_group_config = AWSGroupConfiguration.construct_from_jdict(jdict.get('aws-group-config'))
        aws_project = Project.construct_from_jdict(jdict.get('aws-project'))
        data_log = FilesystemReport.construct_from_jdict(jdict.get('data-log'))
        display_conf = DisplayConf.construct_from_jdict(jdict.get('display-conf'))
        vcal_baseline = VCalBaseline.construct_from_jdict(jdict.get('vcal-baseline'))
        gas_baseline = GasBaseline.construct_from_jdict(jdict.get('gas-baseline'))
        gas_model_conf = GasModelConf.construct_from_jdict(jdict.get('gas-model-conf'))
        gps_conf = GPSConf.construct_from_jdict(jdict.get('gps-conf'))
        interface_conf = InterfaceConf.construct_from_jdict(jdict.get('interface-conf'))
        mpl115a2_calib = MPL115A2Calib.construct_from_jdict(jdict.get('mpl115a2-calib'))
        opc_conf = OPCConf.construct_from_jdict(jdict.get('opc-conf'))
        opc_version = OPCVersion.construct_from_jdict(jdict.get('opc-version'))
        opc_error_summary = OPCErrorSummary.construct_from_jdict(jdict.get('opc-errors'))
        pmx_model_conf = PMxModelConf.construct_from_jdict(jdict.get('pmx-model-conf'))
        pressure_conf = PressureConf.construct_from_jdict(jdict.get('pressure-conf'))
        psu_conf = PSUConf.construct_from_jdict(jdict.get('psu-conf'))
        psu_version = PSUVersion.construct_from_jdict(jdict.get('psu-version'))
        scd30_baseline = SCD30Baseline.construct_from_jdict(jdict.get('scd30-baseline'))
        scd30_conf = SCD30Conf.construct_from_jdict(jdict.get('scd30-conf'))
        schedule = Schedule.construct_from_jdict(jdict.get('schedule'))
        sht_conf = SHTConf.construct_from_jdict(jdict.get('sht-conf'))
        networks = Networks.construct_from_jdict(jdict.get('networks'))
        modem = Modem.construct_from_jdict(jdict.get('modem'))
        sim = SIM.construct_from_jdict(jdict.get('sim'))
        system_id = SystemID.construct_from_jdict(jdict.get('system-id'))
        timezone_conf = TimezoneConf.construct_from_jdict(jdict.get('timezone-conf'))

        return cls(hostname, platform, packs, afe_baseline, afe_id,
                   aws_group_config, aws_project, data_log, display_conf, vcal_baseline,
                   gas_baseline, gas_model_conf, gps_conf, interface_conf, mpl115a2_calib,
                   opc_conf, opc_version, opc_error_summary, pmx_model_conf, pressure_conf,
                   psu_conf, psu_version, scd30_baseline, scd30_conf, schedule,
                   sht_conf, networks, modem, sim, system_id,
                   timezone_conf)


    @classmethod
    def load(cls, manager, psu_version=None, exclude_sim=False):
        csv_logger_conf = CSVLoggerConf.load(manager)

        hostname = socket.gethostname()
        platform = PlatformSummary.construct(manager)
        packs = PackageVersions.construct_from_installation(manager.scs_path(), manager)

        afe_baseline = AFEBaseline.load(manager)
        afe_id = AFEId.load(manager)
        aws_group_config = AWSGroupConfiguration.load(manager)
        aws_project = Project.load(manager)
        data_log = None if csv_logger_conf is None else csv_logger_conf.filesystem_report()
        display_conf = DisplayConf.load(manager)
        vcal_baseline = VCalBaseline.load(manager)
        gas_baseline = GasBaseline.load(manager)
        gas_model_conf = GasModelConf.load(manager)
        gps_conf = GPSConf.load(manager)
        interface_conf = InterfaceConf.load(manager)
        mpl115a2_calib = MPL115A2Calib.load(manager)
        opc_conf = OPCConf.load(manager)
        opc_version = OPCVersion.load(manager)
        opc_error_summary = OPCErrorSummary.load(manager)
        pmx_model_conf = PMxModelConf.load(manager)
        pressure_conf = PressureConf.load(manager)
        psu_conf = PSUConf.load(manager)
        psu_version = psu_version
        scd30_baseline = SCD30Baseline.load(manager)
        scd30_conf = SCD30Conf.load(manager)
        schedule = Schedule.load(manager)
        sht_conf = SHTConf.load(manager)
        networks = None                                         # now provided by Status class
        modem = manager.modem()
        sim = None if exclude_sim else manager.sim()
        system_id = SystemID.load(manager)
        timezone_conf = TimezoneConf.load(manager)

        return cls(hostname, platform, packs, afe_baseline, afe_id,
                   aws_group_config, aws_project, data_log, display_conf, vcal_baseline,
                   gas_baseline, gas_model_conf, gps_conf, interface_conf, mpl115a2_calib,
                   opc_conf, opc_version, opc_error_summary, pmx_model_conf, pressure_conf,
                   psu_conf, psu_version, scd30_baseline, scd30_conf, schedule,
                   sht_conf, networks, modem, sim, system_id,
                   timezone_conf)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, hostname, platform, packs, afe_baseline, afe_id,
                 aws_group_config, aws_project, data_log, display_conf, vcal_baseline,
                 gas_baseline, gas_model_conf, gps_conf, interface_conf, mpl115a2_calib,
                 opc_conf, opc_version, opc_error_summary, pmx_model_conf, pressure_conf,
                 psu_conf, psu_version, scd30_baseline, scd30_conf, schedule,
                 sht_conf, networks, modem, sim, system_id,
                 timezone_conf):
        """
        Constructor
        """

        self.__hostname = hostname                                  # string
        self.__platform = platform                                  # PlatformSummary
        self.__packs = packs                                        # PackageVersions

        self.__afe_baseline = afe_baseline                          # AFEBaseline
        self.__afe_id = afe_id                                      # AFEId
        self.__aws_group_config = aws_group_config                  # AWSGroupConfiguration
        self.__aws_project = aws_project                            # Project
        self.__data_log = data_log                                  # FilesystemReport
        self.__display_conf = display_conf                          # DisplayConf
        self.__vcal_baseline = vcal_baseline                        # VCalBaseline
        self.__gas_baseline = gas_baseline                          # GasBaseline
        self.__gas_model_conf = gas_model_conf                      # GasModelConf
        self.__gps_conf = gps_conf                                  # GPSConf
        self.__interface_conf = interface_conf                      # InterfaceConf
        self.__mpl115a2_calib = mpl115a2_calib                      # MPL115A2Calib
        self.__opc_conf = opc_conf                                  # OPCConf
        self.__opc_version = opc_version                            # OPCVersion
        self.__opc_error_summary = opc_error_summary                # OPCErrorSummary
        self.__pmx_model_conf = pmx_model_conf                      # PMxModelConf
        self.__pressure_conf = pressure_conf                        # PressureConf
        self.__psu_conf = psu_conf                                  # PSUConf
        self.__psu_version = psu_version                            # PSUVersion
        self.__scd30_baseline = scd30_baseline                      # SCD30Baseline
        self.__scd30_conf = scd30_conf                              # SCD30Conf
        self.__schedule = schedule                                  # Schedule
        self.__sht_conf = sht_conf                                  # SHTConf
        self.__networks = networks                                  # Networks or None
        self.__modem = modem                                        # Modem
        self.__sim = sim                                            # SIM
        self.__system_id = system_id                                # SystemID
        self.__timezone_conf = timezone_conf                        # TimezoneConf


    def __eq__(self, other):
        try:
            return self.hostname == other.hostname and self.platform == other.platform and \
                self.packs == other.packs and self.afe_baseline == other.afe_baseline and \
                self.afe_id == other.afe_id and self.aws_project == other.aws_project and \
                self.data_log == other.data_log and self.display_conf == other.display_conf and \
                self.vcal_baseline == other.vcal_baseline and self.gas_baseline == other.gas_baseline and \
                self.gas_model_conf == other.gas_model_conf and self.gps_conf == other.gps_conf and \
                self.interface_conf == other.interface_conf and self.mpl115a2_calib == other.mpl115a2_calib and \
                self.opc_conf == other.opc_conf and self.opc_version == other.opc_version and \
                self.opc_error_summary == other.opc_error_summary and self.pmx_model_conf == other.pmx_model_conf and \
                self.pressure_conf == other.pressure_conf and self.psu_conf == other.psu_conf and \
                self.psu_version == other.psu_version and self.scd30_baseline == other.scd30_baseline and \
                self.scd30_conf == other.scd30_conf and self.schedule == other.schedule and \
                self.sht_conf == other.sht_conf and self.networks == other.networks and \
                self.modem == other.modem and self.sim == other.sim and \
                self.system_id == other.system_id and self.timezone_conf == other.timezone_conf

        except (TypeError, AttributeError):
            return False


    def diff(self, other):
        diff = Configuration(None, None, None, None, None,
                             None, None, None, None, None,
                             None, None, None, None, None,
                             None, None, None, None, None,
                             None, None, None, None, None,
                             None, None, None, None, None,
                             None)

        if self.hostname != other.hostname:
            diff.__hostname = self.hostname

        if self.platform != other.platform:
            diff.__platform = self.platform

        if self.packs != other.packs:
            diff.__packs = self.packs

        if self.afe_id != other.afe_id:
            diff.__afe_id = self.afe_id

        if self.aws_group_config != other.aws_group_config:
            diff.__aws_group_config = self.aws_group_config

        if self.aws_project != other.aws_project:
            diff.__aws_project = self.aws_project

        if self.data_log != other.data_log:
            diff.__data_log = self.data_log

        if self.display_conf != other.display_conf:
            diff.__display_conf = self.display_conf

        if self.vcal_baseline != other.vcal_baseline:
            diff.__vcal_baseline = self.vcal_baseline

        if self.gas_baseline != other.gas_baseline:
            diff.__gas_baseline = self.gas_baseline

        if self.gas_model_conf != other.gas_model_conf:
            diff.__gas_model_conf = self.gas_model_conf

        if self.gps_conf != other.gps_conf:
            diff.__gps_conf = self.gps_conf

        if self.interface_conf != other.interface_conf:
            diff.__interface_conf = self.interface_conf

        if self.mpl115a2_calib != other.mpl115a2_calib:
            diff.__mpl115a2_calib = self.mpl115a2_calib

        if self.opc_conf != other.opc_conf:
            diff.__opc_conf = self.opc_conf

        if self.opc_version != other.opc_version:
            diff.__opc_version = self.opc_version

        if self.pmx_model_conf != other.pmx_model_conf:
            diff.__pmx_model_conf = self.pmx_model_conf

        if self.pressure_conf != other.pressure_conf:
            diff.__pressure_conf = self.pressure_conf

        if self.psu_conf != other.psu_conf:
            diff.__psu_conf = self.psu_conf

        if self.psu_version != other.psu_version:
            diff.__psu_version = self.psu_version

        if self.scd30_baseline != other.scd30_baseline:
            diff.__scd30_baseline = self.scd30_baseline

        if self.scd30_conf != other.scd30_conf:
            diff.__scd30_conf = self.scd30_conf

        if self.schedule != other.schedule:
            diff.__schedule = self.schedule

        if self.sht_conf != other.sht_conf:
            diff.__sht_conf = self.sht_conf

        if self.networks != other.networks:
            diff.__networks = self.networks

        if self.modem != other.modem:
            diff.__modem = self.modem

        if self.sim != other.sim:
            diff.__sim = self.sim

        if self.system_id != other.system_id:
            diff.__system_id = self.system_id

        if self.timezone_conf != other.timezone_conf:
            diff.__timezone_conf = self.timezone_conf

        return diff


    # ----------------------------------------------------------------------------------------------------------------

    def save(self, manager):
        if self.hostname:
            raise ValueError('hostname may not be set')

        if self.platform:
            raise ValueError('platform may not be set')

        if self.packs:
            raise ValueError('packs may not be set')

        if self.afe_baseline:
            self.afe_baseline.save(manager)

        if self.afe_id:
            raise ValueError('afe_id may not be set')

        if self.aws_group_config:
            self.aws_group_config.save(manager)

        if self.aws_project:
            self.aws_project.save(manager)

        if self.data_log:
            raise ValueError('data_log may not be set')

        if self.display_conf:
            self.display_conf.save(manager)

        if self.vcal_baseline:
            self.vcal_baseline.save(manager)

        if self.gas_baseline:
            self.gas_baseline.save(manager)

        if self.gas_model_conf:
            self.gas_model_conf.save(manager)

        if self.gps_conf:
            self.gps_conf.save(manager)

        if self.interface_conf:
            self.interface_conf.save(manager)

        if self.mpl115a2_calib:
            self.mpl115a2_calib.save(manager)

        if self.opc_conf:
            self.opc_conf.save(manager)

        if self.opc_version:
            raise ValueError('opc_version may not be set')

        if self.opc_error_summary:
            raise ValueError('opc_error_summary may not be set')

        if self.pmx_model_conf:
            self.pmx_model_conf.save(manager)

        if self.pressure_conf:
            self.pressure_conf.save(manager)

        if self.psu_conf:
            self.psu_conf.save(manager)

        if self.psu_version:
            raise ValueError('psu_version may not be set')

        if self.scd30_baseline:
            self.scd30_baseline.save(manager)

        if self.scd30_conf:
            self.scd30_conf.save(manager)

        if self.schedule:
            self.schedule.save(manager)

        if self.sht_conf:
            self.sht_conf.save(manager)

        if self.networks:
            raise ValueError('networks may not be set')

        if self.modem:
            raise ValueError('modem may not be set')

        if self.sim:
            raise ValueError('sim may not be set')

        if self.system_id:
            self.system_id.save(manager)

        if self.timezone_conf:
            self.timezone_conf.save(manager)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['hostname'] = self.hostname
        jdict['platform'] = self.platform
        jdict['packs'] = self.packs

        jdict['afe-baseline'] = self.afe_baseline
        jdict['afe-id'] = self.afe_id
        jdict['aws-group-config'] = self.aws_group_config
        jdict['aws-project'] = self.aws_project
        jdict['data-log'] = self.data_log
        jdict['display-conf'] = self.display_conf
        jdict['vcal-baseline'] = self.vcal_baseline
        jdict['gas-baseline'] = self.gas_baseline
        jdict['gas-model-conf'] = self.gas_model_conf
        jdict['gps-conf'] = self.gps_conf
        jdict['interface-conf'] = self.interface_conf
        jdict['mpl115a2-calib'] = self.mpl115a2_calib
        jdict['opc-conf'] = self.opc_conf
        jdict['opc-version'] = self.opc_version
        jdict['opc-errors'] = self.opc_error_summary
        jdict['pmx-model-conf'] = self.pmx_model_conf
        jdict['pressure-conf'] = self.pressure_conf
        jdict['psu-conf'] = self.psu_conf
        jdict['psu-version'] = self.psu_version
        jdict['scd30-baseline'] = self.scd30_baseline
        jdict['scd30-conf'] = self.scd30_conf
        jdict['schedule'] = self.schedule
        jdict['sht-conf'] = self.sht_conf
        jdict['networks'] = self.networks
        jdict['modem'] = self.modem
        jdict['sim'] = self.sim
        jdict['system-id'] = self.system_id
        jdict['timezone-conf'] = self.timezone_conf

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def hostname(self):
        return self.__hostname


    @property
    def platform(self):
        return self.__platform


    @property
    def packs(self):
        return self.__packs


    @property
    def afe_baseline(self):
        return self.__afe_baseline


    @property
    def afe_id(self):
        return self.__afe_id


    @property
    def aws_group_config(self):
        return self.__aws_group_config


    @property
    def aws_project(self):
        return self.__aws_project


    @property
    def data_log(self):
        return self.__data_log


    @property
    def display_conf(self):
        return self.__display_conf


    @property
    def vcal_baseline(self):
        return self.__vcal_baseline


    @property
    def gas_baseline(self):
        return self.__gas_baseline


    @property
    def gas_model_conf(self):
        return self.__gas_model_conf


    @property
    def gps_conf(self):
        return self.__gps_conf


    @property
    def interface_conf(self):
        return self.__interface_conf


    @property
    def mpl115a2_calib(self):
        return self.__mpl115a2_calib


    @property
    def opc_conf(self):
        return self.__opc_conf


    @property
    def opc_version(self):
        return self.__opc_version


    @property
    def opc_error_summary(self):
        return self.__opc_error_summary


    @property
    def pmx_model_conf(self):
        return self.__pmx_model_conf


    @property
    def pressure_conf(self):
        return self.__pressure_conf


    @property
    def psu_conf(self):
        return self.__psu_conf


    @property
    def psu_version(self):
        return self.__psu_version


    @property
    def scd30_baseline(self):
        return self.__scd30_baseline


    @property
    def scd30_conf(self):
        return self.__scd30_conf


    @property
    def schedule(self):
        return self.__schedule


    @property
    def sht_conf(self):
        return self.__sht_conf


    @property
    def networks(self):
        return self.__networks


    @property
    def modem(self):
        return self.__modem


    @property
    def sim(self):
        return self.__sim


    @property
    def system_id(self):
        return self.__system_id


    @property
    def timezone_conf(self):
        return self.__timezone_conf


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Configuration:{hostname:%s, platform:%s, packs:%s, afe_baseline:%s, afe_id:%s, " \
               "aws_group_config:%s, aws_project:%s, data_log:%s, display_conf:%s, vcal_baseline:%s, " \
               "gas_baseline:%s, gas_model_conf:%s, gps_conf:%s, interface_conf:%s, mpl115a2_calib:%s, " \
               "opc_conf:%s, opc_version:%s, opc_error_summary:%s, pmx_model_conf:%s, pressure_conf:%s, " \
               "psu_conf:%s, psu_version:%s, scd30_baseline:%s, scd30_conf:%s, schedule:%s, " \
               "sht_conf:%s, networks:%s, modem:%s, sim:%s, system_id:%s, " \
               "timezone_conf:%s}" % \
               (self.hostname, self.platform, self.packs, self.afe_baseline, self.afe_id,
                self.aws_group_config, self.aws_project, self.data_log, self.display_conf, self.vcal_baseline,
                self.gas_baseline, self.gas_model_conf, self.gps_conf, self.interface_conf, self.mpl115a2_calib,
                self.opc_conf, self.opc_version, self.opc_error_summary, self.pmx_model_conf, self.pressure_conf,
                self.psu_conf, self.psu_version, self.scd30_baseline, self.scd30_conf, self.schedule,
                self.sht_conf, self.networks, self.modem, self.sim, self.system_id,
                self.timezone_conf)



"""
Created on 17 Apr 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://pythoncentral.io/hashing-strings-with-python/

example document:
{"tag": "my-laptop", "attn": "scs-ap1-6", "rec": "2022-11-28T12:05:26Z", "ver": 2.0,
"cmd_tokens": ["test"], "timeout": 20, "digest": "183f9036b2afd0f347ae30f20b228f962ba9731d"}
"""

import hashlib

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable, JSONify

from scs_core.sample.sample import Sample


# --------------------------------------------------------------------------------------------------------------------

class ControlDatum(JSONable):
    """
    classdocs
    """

    VERSION = 1.0                   # Version 2.0 is not compatible with deployed devices

    __DEFAULT_TIMEOUT = 30.0        # seconds

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        tag = jdict.get('tag')
        attn = jdict.get('attn')
        rec = LocalizedDatetime.construct_from_iso8601(jdict.get('rec'))

        try:
            version = round(float(jdict.get('ver')), 1)
        except (TypeError, ValueError):
            version = None

        cmd_tokens = jdict.get('cmd_tokens')
        timeout = jdict.get('timeout', cls.__DEFAULT_TIMEOUT)
        digest = jdict.get('digest')

        datum = cls(tag, attn, rec, cmd_tokens, timeout, digest, version=version)

        return datum


    @classmethod
    def construct(cls, tag, attn, rec, cmd_tokens, timeout, key):
        digest = cls.__hash(tag, attn, rec, cmd_tokens, key, cls.VERSION)

        return cls(tag, attn, rec, cmd_tokens, timeout, digest, version=cls.VERSION)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def __hash(cls, tag, attn, rec, cmd_tokens, key, version):
        rec_iso8601 = rec.as_iso8601(include_millis=Sample.INCLUDE_MILLIS)
        text = str(tag) + str(attn) + JSONify.dumps(rec_iso8601) + str(cmd_tokens) + str(key)

        if version == 2.0:
            hash_object = hashlib.sha1(text.encode())
        else:
            hash_object = hashlib.sha256(text.encode())

        return hash_object.hexdigest()


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag, attn, rec, cmd_tokens, timeout, digest, version=None):
        """
        Constructor
        """
        self.__tag = tag                    # string - originator of message
        self.__attn = attn                  # string - intended recipient of message
        self.__rec = rec                    # LocalizedDatetime

        self.__version = version            # float

        self.__cmd_tokens = cmd_tokens      # array of { string | int | float }
        self.__timeout = int(timeout)       # int
        self.__digest = digest              # string


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self, key):
        digest = self.__hash(self.tag, self.attn, self.rec, self.cmd_tokens, key, self.version)

        return digest == self.digest


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['tag'] = self.tag
        jdict['attn'] = self.attn
        jdict['rec'] = self.rec.as_iso8601(include_millis=Sample.INCLUDE_MILLIS)

        jdict['ver'] = round(self.version, 1)

        jdict['cmd_tokens'] = self.cmd_tokens
        jdict['timeout'] = self.timeout
        jdict['digest'] = self.digest

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def tag(self):
        return self.__tag


    @property
    def attn(self):
        return self.__attn


    @property
    def rec(self):
        return self.__rec


    @property
    def version(self):
        return self.__version


    @property
    def cmd_tokens(self):
        return self.__cmd_tokens


    @property
    def timeout(self):
        return self.__timeout


    @property
    def digest(self):
        return self.__digest


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ControlDatum:{tag:%s, attn:%s, rec:%s, version:%s, cmd_tokens:%s, timeout:%s, digest:%s}" % \
               (self.tag, self.attn, self.rec, self.version, self.cmd_tokens, self.timeout, self.digest)



"""
Created on 20 Jan 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://realpython.com/python-logging/
"""

import logging
import sys


# --------------------------------------------------------------------------------------------------------------------

class LoggingSpecification(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, name, level):
        """
        Constructor
        """
        self.__name = name                              # string
        self.__level = level                            # int or string


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def name(self):
        return self.__name


    @property
    def level(self):
        return self.__level


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "LoggingSpecification:{name:%s, level:%s}" %   (self.name, self.level)


# --------------------------------------------------------------------------------------------------------------------

# noinspection PyPep8Naming
class Logging(object):
    """
    classdocs
    """

    __NAME = None
    __LEVEL = logging.NOTSET

    __MULTI_FORMAT = '%(name)s: %(message)s'

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def config(cls, name, verbose=False, level=logging.ERROR, stream=sys.stderr):       # for CLU
        cls.__NAME = name
        cls.__LEVEL = logging.INFO if verbose else level

        logging.basicConfig(format=cls.__MULTI_FORMAT, level=cls.__LEVEL, stream=stream)


    @classmethod
    def replicate(cls, specification: LoggingSpecification, stream=sys.stderr):         # for child process
        cls.__NAME = specification.name
        cls.__LEVEL = specification.level

        logging.basicConfig(format=cls.__MULTI_FORMAT, level=cls.__LEVEL, stream=stream)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def getLogger(cls, name=None):
        logger_name = cls.__NAME if cls.__NAME else name

        return logging.getLogger(name=logger_name)


    @classmethod
    def debugging_on(cls):
        return cls.__LEVEL == logging.DEBUG


    @classmethod
    def specification(cls):
        return LoggingSpecification(cls.__NAME, cls.__LEVEL)


    @classmethod
    def level(cls):
        return cls.__LEVEL



"""
Created on 17 Jun 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint
from scs_core.aws.monitor.device.device_monitor_report import DeviceMonitorReport


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('DevMonStatAPI/DeviceMonitorStatus',
                   'https://0l7fwqkzr8.execute-api.us-west-2.amazonaws.com/default/DeviceMonitorStatus')


# --------------------------------------------------------------------------------------------------------------------

class DeviceMonitorStatusManager(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def find(self, token, device_tag_filter=None, exact=False):
        payload = {
            'device_tag': device_tag_filter,
            'exact': exact
        }

        response = requests.get(Endpoint.url(), headers=self._token_headers(token), json=payload)
        self._check_response(response)

        return DeviceMonitorReport.construct_from_jdict(response.json())



"""
Created on 15 Sep 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://pymotw.com/2/subprocess/
"""

import json
import sys

from subprocess import Popen, PIPE

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class JSONPopen(Popen):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, args, bufsize=-1, executable=None,
                 stdin=None, stdout=None, stderr=None,
                 preexec_fn=None, close_fds=True,
                 shell=False, cwd=None, env=None, universal_newlines=None,
                 startupinfo=None, creationflags=0,
                 restore_signals=True, start_new_session=False,
                 pass_fds=(), encoding=None, errors=None, text=None):
        """
        Constructor
        """
        super().__init__(args, bufsize=bufsize, executable=executable,
                         stdin=stdin, stdout=stdout, stderr=stderr,
                         preexec_fn=preexec_fn, close_fds=close_fds,
                         shell=shell, cwd=cwd, env=env, universal_newlines=universal_newlines,
                         startupinfo=startupinfo, creationflags=creationflags,
                         restore_signals=restore_signals, start_new_session=start_new_session,
                         pass_fds=pass_fds, encoding=encoding, errors=errors, text=text)


    # ----------------------------------------------------------------------------------------------------------------

    def json(self):
        return json.loads(self.stdout.readline().decode())


    def json_lines(self):
        for line in self.stdout.readlines():
            yield json.loads(line.decode())


# --------------------------------------------------------------------------------------------------------------------

class Command(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, verbose=False, on_abort=None):
        """
        Constructor
        """
        self.__verbose = verbose                            # bool
        self.__on_abort = on_abort                          # function

        self.__logger = Logging.getLogger()


    # ----------------------------------------------------------------------------------------------------------------

    def s(self, cmd_args, wait=True, no_verbose=False, abort_on_fail=True):
        tokens = self.__cmd(cmd_args, no_verbose=no_verbose)
        p = JSONPopen(' '.join(tokens), shell=True)

        return self.__process(p, wait=wait, abort_on_fail=abort_on_fail)


    def o(self, cmd_args, wait=False, no_verbose=False, abort_on_fail=True):
        tokens = self.__cmd(cmd_args, no_verbose=no_verbose)
        p = JSONPopen(tokens, stdout=PIPE)

        return self.__process(p, wait=wait, abort_on_fail=abort_on_fail)


    def io(self, p: JSONPopen, cmd_args, wait=False, no_verbose=False, abort_on_fail=True):
        tokens = self.__cmd(cmd_args, no_verbose=no_verbose)
        p = JSONPopen(tokens, stdin=p.stdout, stdout=PIPE)

        return self.__process(p, wait=wait, abort_on_fail=abort_on_fail)


    def i(self, p: JSONPopen, cmd_args, wait=True, no_verbose=False, abort_on_fail=True):
        tokens = self.__cmd(cmd_args, no_verbose=no_verbose)
        p = JSONPopen(tokens, stdin=p.stdout, stdout=sys.stderr)

        return self.__process(p, wait=wait, abort_on_fail=abort_on_fail)


    # ----------------------------------------------------------------------------------------------------------------

    def __cmd(self, cmd_args, no_verbose=False):
        tokens = [str(cmd_arg) for cmd_arg in cmd_args if cmd_arg is not None]

        if self.__verbose and not no_verbose:
            tokens = tokens[:1] + ['-v'] + tokens[1:]

        if self.__verbose:                          # Command verbosity overrides logger verbosity
            self.__logger.info(' '.join(tokens))

        return tokens


    def __process(self, p: JSONPopen, wait=False, abort_on_fail=True):
        if wait:
            p.wait()

        if abort_on_fail and (p.returncode is not None and p.returncode != 0):
            self.abort(p.returncode)

        return p


    def abort(self, return_code):
        self.__logger.error('ABORTED.')

        if self.__on_abort:
            self.__on_abort()

        exit(return_code)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def verbose(self):
        return self.__verbose


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Command:{verbose:%s, on_abort:%s}" % (self.verbose, self.__on_abort.__class__.__name__)



"""
Created on 11 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
{"topic": "particulates", "species": "pm1", "site-code": null}
"""

from collections import OrderedDict

from scs_core.aqcsv.conf.airnow_site_conf import AirNowSiteConf

from scs_core.aqcsv.connector.source_mapping import SourceMapping

from scs_core.aqcsv.data.aqcsv_datetime import AQCSVDatetime
from scs_core.aqcsv.data.aqcsv_record import AQCSVRecord

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable
from scs_core.data.path_dict import PathDict

from scs_core.location.timezone import Timezone

from scs_core.position.gps_datum import GPSDatum


# --------------------------------------------------------------------------------------------------------------------

class DatumMapping(JSONable):
    """
    classdocs
    """

    __SCHEDULES = {
        'gases': 'scs-gases',
        'particulates': 'scs-particulates'
    }

    @classmethod
    def is_valid_topic(cls, topic):
        return topic in cls.__SCHEDULES


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if jdict is None:
            return None

        topic = jdict.get('topic')
        species = jdict.get('species')
        site_code = jdict.get('site-code')

        return DatumMapping(topic, species, site_code)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, topic, species, site_code=None):
        """
        Constructor
        """
        self.__topic = topic                                        # string
        self.__species = species                                    # string
        self.__site_code = site_code                                # string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['topic'] = self.topic
        jdict['species'] = self.species
        jdict['site-code'] = self.site_code

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def aqcsv_record(self, datum: PathDict, duration):
        # validation...
        if self.environment_tag(datum) != self.status_tag(datum):
            raise ValueError("non-matching tag fields: %s" % datum)

        # parameter_code...
        mapping = self.aqcsv_source_mapping(datum)
        parameter_code = mapping.parameter_code

        # site_code / POCs...
        if self.__site_code is not None:
            code = self.__site_code
            poc = 1

        else:
            site_conf = self.site_conf(datum)
            code = site_conf.site.as_code()
            poc = site_conf.poc(parameter_code)

        # datetime_code...
        rec = self.aqcsv_rec(datum)

        # position...
        gps = self.gps(datum)

        if gps is not None and gps.elv is not None:
            lat = gps.pos.lat
            lon = gps.pos.lng
            gis_datum = AQCSVRecord.GIS_DATUM
            elev = int(round(gps.elv))

        else:
            lat = None
            lon = None
            gis_datum = None
            elev = None

        # record...
        record = AQCSVRecord(
            site_code=code,

            data_status=AQCSVRecord.STATUS_FINAL,
            action_code=AQCSVRecord.ACTION_DEFAULT,

            datetime_code=rec.as_json(),

            parameter_code=parameter_code,

            duration=duration,
            frequency=0,

            value=self.value(datum),
            unit_code=mapping.unit_code,

            qc_code=mapping.qc_code,
            poc=poc,

            lat=lat,
            lon=lon,
            gis_datum=gis_datum,
            elev=elev,

            method_code=mapping.method_code,
            mpc_code=mapping.mpc_code,
            mpc_value=mapping.mpc_value)

        return record


    # ----------------------------------------------------------------------------------------------------------------
    # environment fields...

    def environment_tag(self, datum: PathDict):
        tag_path = '.'.join([self.topic, 'tag'])

        return datum.node(tag_path)


    def aqcsv_rec(self, datum: PathDict):
        localised = LocalizedDatetime.construct_from_jdict(datum.node('rec'))
        timezone = self.timezone(datum)

        return AQCSVDatetime(localised.datetime, timezone.zone)


    def aqcsv_source_mapping(self, datum: PathDict):
        pk = (self.topic, self.species, self.source(datum))

        mapping = SourceMapping.instance(pk)

        if mapping is None:
            raise KeyError("no source mapping found for %s" % str(pk))

        return mapping


    def value(self, datum: PathDict):
        species_path = '.'.join([self.topic, 'val', self.species])

        return datum.node(species_path)


    def source(self, datum: PathDict):
        source_path = '.'.join([self.topic, 'src'])

        return datum.node(source_path)


    # ----------------------------------------------------------------------------------------------------------------
    # status fields...

    @classmethod
    def status_tag(cls, datum: PathDict):
        return datum.node('status.tag')


    @classmethod
    def site_conf(cls, datum: PathDict):
        jdict = datum.node('status.val.airnow')

        conf = AirNowSiteConf.construct_from_jdict(jdict)

        if conf is None:
            raise KeyError("no site configuration found for %s" % str(jdict))

        return conf


    @classmethod
    def gps(cls, datum: PathDict):
        jdict = datum.node('status.val.gps')

        return GPSDatum.construct_from_jdict(jdict)


    @classmethod
    def timezone(cls, datum: PathDict):
        jdict = datum.node('status.val.tz')

        return Timezone.construct_from_jdict(jdict)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def topic(self):
        return self.__topic


    @property
    def species(self):
        return self.__species


    @property
    def site_code(self):
        return self.__site_code


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "DatumMapping:{topic:%s, species:%s, site_code:%s}" % (self.topic, self.species, self.site_code)



"""
Created on 11 Jul 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import time


# --------------------------------------------------------------------------------------------------------------------

class IntervalTimer(object):
    """
    generate indices at pre-set intervals
    """

    def __init__(self, interval):
        self.__interval = interval
        self.__next_yield = time.time() + self.__interval


    # ----------------------------------------------------------------------------------------------------------------

    def range(self, stop):
        for i in range(stop):
            try:
                self.__sleep_until_next_yield()
            except KeyboardInterrupt:
                return stop

            yield i


    def true(self):
        try:
            self.__sleep_until_next_yield()
        except KeyboardInterrupt:
            return False

        return True


    def reset(self):
        self.__next_yield = time.time() + self.__interval


    # ----------------------------------------------------------------------------------------------------------------

    def __sleep_until_next_yield(self):
        if self.__interval == 0:
            return

        sleep_time = (self.__next_yield - time.time()) % self.__interval        # prevents negative intervals

        if sleep_time == 0:
            sleep_time = self.__interval        # prevents zero intervals

        time.sleep(sleep_time)

        self.__next_yield += self.__interval


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def interval(self):
        return self.__interval


    @property
    def time_to_next_yield(self):
        return abs(self.__next_yield - time.time())         # time to next must always be positive


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "IntervalTimer:{interval:%s, time_to_next_yield:%s}" % \
               (round(self.interval, 1), round(self.time_to_next_yield, 1))



"""
Created on 2 Feb 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
{"account": {"username": "scs-bgx-867", "invoice": "INV-0489", "created": "2023-08-30T13:52:28Z",
"last-updated": "2023-08-30T13:52:29Z"}, "memberships": [
{"DeviceTag": "scs-bgx-867", "OrgID": 16, "DeploymentLabel": "Development 19",
"DevicePath": "cirrusresearch/development/device/praxis-000867/", "LocationPath": "cirrusresearch/development/loc/19/",
"StartDatetime": "1970-01-01T00:00:00Z", "EndDatetime": null}]}
"""

from collections import OrderedDict

from scs_core.data.array_dict import ArrayDict
from scs_core.data.json import JSONable
from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class CognitoMembership(JSONable):
    """
    classdocs
    """

    @classmethod
    def merge(cls, cognito_accounts, org_memberships):
        org_dict = ArrayDict([(org_membership.username, org_membership) for org_membership in sorted(org_memberships)])

        # Accounts...
        return [cls(cognito_account, org_dict.get(cognito_account.username)) for cognito_account in cognito_accounts]


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, cognito_account, memberships):
        """
        Constructor
        """
        self.__cognito_account = cognito_account            # CognitoUserIdentity or CognitoDeviceIdentity
        self.__memberships = memberships                    # array of OrganisationUser


    def __lt__(self, other):
        return self.cognito_account < other.cognito_account


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['account'] = self.cognito_account
        jdict['memberships'] = self.memberships

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def cognito_account(self):
        return self.__cognito_account


    @property
    def memberships(self):
        return self.__memberships


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CognitoMembership:{cognito_account:%s, memberships:%s}" % \
               (self.cognito_account, Str.collection(self.memberships))



"""
Created on 16 Jul 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

specifies which GPS receiver is present, if any, plus sample interval, tally and verbosity

example JSON:
{"model": "PAM7Q", "sample-interval": 10, "tally": 1, "report-file": "/tmp/southcoastscience/gps_report.json",
"debug": false}
"""

from collections import OrderedDict

from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class GPSConf(PersistentJSONable):
    """
    classdocs
    """

    DEFAULT_SAMPLE_INTERVAL =       10          # seconds
    DEFAULT_TALLY =                 60          # 10 minutes


    __FILENAME = "gps_conf.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        model = jdict.get('model')

        sample_interval = jdict.get('sample-interval', cls.DEFAULT_SAMPLE_INTERVAL)
        tally = jdict.get('tally', cls.DEFAULT_TALLY)
        report_file = jdict.get('report-file')
        debug = jdict.get('debug', False)

        return cls(model, sample_interval, tally, report_file, debug)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, model, sample_interval, tally, report_file, debug):
        """
        Constructor
        """
        super().__init__()

        self.__model = model                                        # string

        self.__sample_interval = int(sample_interval)               # int seconds
        self.__tally = int(tally)                                   # int count
        self.__report_file = report_file                            # string tmp file to store current GPS report
        self.__debug = bool(debug)                                  # bool


    def __eq__(self, other):
        try:
            return self.model == other.model and self.sample_interval == other.sample_interval and \
                   self.tally == other.tally and self.report_file == other.report_file and \
                   self.debug == other.debug

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def gps_monitor(self, interface, host):
        return None


    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def gps(self, interface, host):
        return None


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def model(self):
        return self.__model


    @property
    def sample_interval(self):
        return self.__sample_interval


    @property
    def tally(self):
        return self.__tally


    @property
    def report_file(self):
        return self.__report_file


    @property
    def debug(self):
        return self.__debug


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['model'] = self.model

        jdict['sample-interval'] = self.sample_interval
        jdict['tally'] = self.tally
        jdict['report-file'] = self.report_file
        jdict['debug'] = self.debug

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "GPSConf(core):{model:%s, sample_interval:%s, tally:%s, report_file:%s, debug:%s}" % \
               (self.model, self.sample_interval, self.tally, self.report_file, self.debug)



"""
Created on 11 Aug 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

A JSONable extension to pytz.timezone.

https://stackoverflow.com/questions/13866926/python-pytz-list-of-timezones

example JSON:
{"name": "US/Michigan", "utc-offset": "-04:00"}
"""

import datetime
import pytz

from collections import OrderedDict

from scs_core.data.json import JSONable
from scs_core.location.timezone_offset import TimezoneOffset


# --------------------------------------------------------------------------------------------------------------------

class Timezone(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def is_valid(cls, name):
        return name in pytz.all_timezones


    @classmethod
    def zones(cls):
        return pytz.all_timezones


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        name = jdict.get('name')

        return Timezone(name)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, name):
        """
        Constructor
        """
        if not Timezone.is_valid(name):
            raise ValueError("Invalid timezone name: %s" % name)

        self.__name = name                              # a Pytz timezone name


    # ----------------------------------------------------------------------------------------------------------------

    def utc_offset(self, year, month, day, hour):
        tz = pytz.timezone(self.__name)
        dt = datetime.datetime(year=year, month=month, day=day, hour=hour)

        offset = tz.localize(dt).strftime('%z')

        return TimezoneOffset.construct_from_offset(offset)


    def current_utc_offset(self):
        if self.name is None:
            return None

        tz = pytz.timezone(self.name)
        dt = datetime.datetime.now(tz)

        offset = dt.strftime('%z')

        return TimezoneOffset.construct_from_offset(offset)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['name'] = self.name
        jdict['utc-offset'] = self.current_utc_offset()

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def name(self):
        return self.__name


    @property
    def zone(self):
        return pytz.timezone(self.__name)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Timezone:{name:%s}" % self.name



"""
Created on 30 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Alphasense A4 photo ionisation detector

PID-AH sn: 143950149 - pid_elc_mv: 47.6, pid_sens_mv_ppm: 0.0375

PID-AH "default" - pid_elc_mv: 54, pid_sens_mv_ppm: 0.040
"""

from scs_core.gas.pid.pid_calib import PIDCalib
from scs_core.gas.pid.pid_calibrated_datum import PIDCalibrator
from scs_core.gas.pid.pid_datum import PIDDatum
from scs_core.gas.pid.pid_temp_comp import PIDTempComp

from scs_core.gas.sensor import Sensor


# --------------------------------------------------------------------------------------------------------------------

class PID(Sensor):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------
    #

    @classmethod
    def init(cls):
        # ppb sensitivity...
        cls.SENSORS[cls.CODE_VOC_PPB_T1] = PID(cls.CODE_VOC_PPB_T1,  'PIDH2', 'VOC',  4, 50.0, 40.0)    # was PIDNH
        cls.SENSORS[cls.CODE_VOC_PPB_T2] = PID(cls.CODE_VOC_PPB_T2,  'PIDH2', 'VOC',  4, 50.0, 40.0)    # was PIDNH
        cls.SENSORS[cls.CODE_VOC_PPB_T3] = PID(cls.CODE_VOC_PPB_T2,  'PIDH2', 'VOC',  4, 50.0, 40.0)    # was PIDNH
        cls.SENSORS[cls.CODE_VOC_PPB_T4] = PID(cls.CODE_VOC_PPB_T2,  'PIDH2', 'VOC',  4, 50.0, 40.0)    # was PIDNH
        cls.SENSORS[cls.CODE_VOC_PPB_T5] = PID(cls.CODE_VOC_PPB_T2,  'PIDH2', 'VOC',  4, 50.0, 40.0)    # was PIDNH

        # ppm sensitivity
        cls.SENSORS[cls.CODE_VOC_PPM] = PID(cls.CODE_VOC_PPM,  'VOC',  'PID12', 4, 50.0, 0.040)     # was PIDN1


    # ----------------------------------------------------------------------------------------------------------------

    __MINIMUM_CNC = 250              # lowest valid reading (ppb)

    @classmethod
    def sensitivity(cls, correct, reported):
        sensitivity = (reported - cls.__MINIMUM_CNC) / (correct - cls.__MINIMUM_CNC)

        return round(sensitivity, 3)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, sensor_code, sensor_type, gas_name, adc_gain_index, default_elc_mv, default_sens_mv_ppm):
        """
        Constructor
        """
        Sensor.__init__(self, sensor_code, sensor_type, gas_name, adc_gain_index)

        self.__default_elc_mv = default_elc_mv
        self.__default_sens_mv_ppm = default_sens_mv_ppm

        self.__tc = PIDTempComp.find(sensor_code)


    # ----------------------------------------------------------------------------------------------------------------

    def sample(self, afe, temp, sensor_index, no2_sample=None):
        we_v = afe.sample_raw_wrk(sensor_index, self.adc_gain_index)

        datum = PIDDatum.construct(self.calib, self.baseline, self.__tc, temp, we_v)

        if self.calibrator is None:
            return datum

        return self.calibrator.calibrate(datum)


    def null_datum(self):
        return PIDDatum(None)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def default_elc_mv(self):
        return self.__default_elc_mv


    @property
    def default_sens_mv_ppm(self):
        return self.__default_sens_mv_ppm


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def calib(self):
        return self._calib


    @calib.setter
    def calib(self, calib):
        # replace missing values with defaults...
        pid_elc_mv = self.default_elc_mv if calib.pid_elc_mv is None else calib.pid_elc_mv
        pid_sens_mv_ppm = self.default_sens_mv_ppm if calib.pid_sens_mv_ppm is None else calib.pid_sens_mv_ppm

        # set calibration...
        self._calib = PIDCalib(calib.serial_number, calib.sensor_type, pid_elc_mv, pid_sens_mv_ppm, calib.bump_calib)
        self._calibrator = PIDCalibrator(self.calib)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "PID:{sensor_code:%s, sensor_type:%s, gas_name:%s, adc_gain_index:0x%04x, default_elc_mv:%s, " \
               "default_sens_mv_ppm:%s, calib:%s, baseline:%s}" %  \
               (self.sensor_code, self.sensor_type, self.gas_name, self.adc_gain_index, self.default_elc_mv,
                self.default_sens_mv_ppm, self.calib, self.baseline)



"""
Created on 23 Dec 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import json

from abc import abstractmethod

from scs_core.comms.uds_client import UDSClient
from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class PMxInferenceClient(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, uds_client: UDSClient):
        """
        Constructor
        """
        self.__logger = Logging.getLogger()

        self._uds_client = uds_client                       # UDSClient


    # ----------------------------------------------------------------------------------------------------------------

    def wait_for_server(self):
        self._uds_client.open()

        self._uds_client.request(json.dumps(None))
        self._uds_client.wait_for_response()

        self.__logger.info('connected to server')


    def open(self):
        self._uds_client.open()


    def close(self):
        self._uds_client.close()


    def model_name(self):
        self._uds_client.request('"?"')
        response = self._uds_client.wait_for_response()

        return json.loads(response)


    # ----------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def infer(self, opc_sample, ext_sht_datum):
        pass



"""
Created on 26 Nov 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

This version of MessageManager for AWS lambda function with multipart response.
"""

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint
from scs_core.aws.manager.topic_history.topic_history_intercourse import TopicHistoryRequest, TopicHistoryResponse


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('DataAPI/TopicHistory',
                   'https://60rd4rxw81.execute-api.us-west-2.amazonaws.com/default/TopicHistory')


# --------------------------------------------------------------------------------------------------------------------

class TopicHistoryFinder(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, reporter=None):
        super().__init__(reporter=reporter)


    # ----------------------------------------------------------------------------------------------------------------

    def find_latest_for_topic(self, token, topic, end, path, include_wrapper, rec_only, backoff_limit):
        documents = list(self.find_for_topic(token, topic, None, end, path, False, None, include_wrapper, rec_only,
                                             False, False, True, backoff_limit))
        if documents:
            return documents[0]

        return None


    def find_for_topic(self, token, topic, start, end, path, fetch_last, checkpoint, include_wrapper, rec_only,
                       min_max, exclude_remainder, fetch_last_written_before, backoff_limit):
        self._reporter.reset()

        request = TopicHistoryRequest(topic, start, end, path, fetch_last, checkpoint, include_wrapper, rec_only,
                                      min_max, exclude_remainder, fetch_last_written_before, backoff_limit)
        self._logger.debug(request)

        for item in self._get_blocks(Endpoint.url(), token, TopicHistoryResponse, params=request.params()):
            yield item
